/* crypto/asn1/x_xPGP.c */
/*
 * Copyright (C) 2005-2006 Mark Fernie (drbob@lunamutt.com)  
 * All rights reserved.
 */
/* ====================================================================
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 *
 */

#include <stdio.h>
#include "cryptlib.h"
#include <openssl/evp.h>
#include <openssl/asn1t.h>
#include <openssl/x509.h>
#include <openssl/xPGP.h>


ASN1_SEQUENCE(XPGP_UID) = {
	ASN1_SIMPLE(XPGP_UID, subject, X509_NAME),
} ASN1_SEQUENCE_END(XPGP_UID)

IMPLEMENT_ASN1_FUNCTIONS(XPGP_UID)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP_UID)

ASN1_SEQUENCE(XPGP_PUBKEY) = {
	ASN1_SIMPLE(XPGP_PUBKEY, version, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP_PUBKEY, validity, X509_VAL),
	ASN1_SIMPLE(XPGP_PUBKEY, key, X509_PUBKEY)
} ASN1_SEQUENCE_END(XPGP_PUBKEY)

/* No Longer in the PUBKEY.
 *	ASN1_SIMPLE(XPGP_PUBKEY, keytype, X509_ALGOR),
 *	ASN1_SIMPLE(XPGP_PUBKEY, keyid, ASN1_OCTET_STRING),
 */

IMPLEMENT_ASN1_FUNCTIONS(XPGP_PUBKEY)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP_PUBKEY)

ASN1_SEQUENCE(XPGP_TRUST_OBJ) = {
	ASN1_SIMPLE(XPGP_TRUST_OBJ, trust_lvl, ASN1_INTEGER)
} ASN1_SEQUENCE_END(XPGP_TRUST_OBJ)

IMPLEMENT_ASN1_FUNCTIONS(XPGP_TRUST_OBJ)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP_TRUST_OBJ)


ASN1_SEQUENCE(XPGP_SIGNDATA) = {
	ASN1_SIMPLE(XPGP_SIGNDATA, version, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP_SIGNDATA, type, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP_SIGNDATA, timestamp, ASN1_TIME),
	ASN1_SIMPLE(XPGP_SIGNDATA, validity, X509_VAL),
	ASN1_SIMPLE(XPGP_SIGNDATA, key, X509_PUBKEY),
	ASN1_SIMPLE(XPGP_SIGNDATA, subject, X509_NAME),
	ASN1_SIMPLE(XPGP_SIGNDATA, issuer, X509_NAME),
	ASN1_SIMPLE(XPGP_SIGNDATA, signtype, X509_ALGOR)
} ASN1_SEQUENCE_END(XPGP_SIGNDATA)

IMPLEMENT_ASN1_FUNCTIONS(XPGP_SIGNDATA)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP_SIGNDATA)


static int xPGP_SIGNATURE_cb(int operation, ASN1_VALUE **pval, const ASN1_ITEM *it)
{
	XPGP_SIGNATURE *ret = (XPGP_SIGNATURE *)*pval;

	switch(operation) {

		case ASN1_OP_NEW_POST:
		ret->trust_signer = NULL;
		break;

		case ASN1_OP_D2I_POST:
		break;

		case ASN1_OP_FREE_POST:
		if (ret -> trust_signer != NULL)
			XPGP_TRUST_OBJ_free(ret -> trust_signer);

		break;

	}

	return 1;

}

ASN1_SEQUENCE_ref(XPGP_SIGNATURE, xPGP_SIGNATURE_cb, CRYPTO_LOCK_X509) = {
	ASN1_SIMPLE(XPGP_SIGNATURE, version, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP_SIGNATURE, type, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP_SIGNATURE, timestamp, ASN1_TIME),

	ASN1_SIMPLE(XPGP_SIGNATURE, issuer, X509_NAME),
	ASN1_SIMPLE(XPGP_SIGNATURE, keyid, ASN1_OCTET_STRING),

	ASN1_SIMPLE(XPGP_SIGNATURE, signtype, X509_ALGOR),
	ASN1_SIMPLE(XPGP_SIGNATURE, signature, ASN1_BIT_STRING)
} ASN1_SEQUENCE_END_ref(XPGP_SIGNATURE, XPGP_SIGNATURE)

IMPLEMENT_ASN1_FUNCTIONS(XPGP_SIGNATURE)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP_SIGNATURE)

static ASN1_METHOD meth_sig={
	(int (*)())  i2d_XPGP_SIGNATURE,
	(char *(*)())d2i_XPGP_SIGNATURE,
	(char *(*)())XPGP_SIGNATURE_new,
	(void (*)()) XPGP_SIGNATURE_free};

ASN1_METHOD *XPGP_SIGNATURE_asn1_meth(void)
	{
	return(&meth_sig);
	}


/* Callback for XPGP, similar to X509 */

static int xPGP_cb(int operation, ASN1_VALUE **pval, const ASN1_ITEM *it)
{
	XPGP *ret = (XPGP *)*pval;

	switch(operation) {

		case ASN1_OP_NEW_POST:
		ret->valid=0;
		ret->name = NULL;
		ret->trust_to_sign = NULL;
		ret->trust_cert = NULL;

		break;

		case ASN1_OP_D2I_POST:

		if (ret->name != NULL) OPENSSL_free(ret->name);
		ret->name=X509_NAME_oneline(ret->subject->subject,NULL,0);

		break;

		case ASN1_OP_FREE_POST:

		if (ret -> trust_to_sign != NULL)
			XPGP_TRUST_OBJ_free(ret -> trust_to_sign);
		if (ret -> trust_cert != NULL)
			XPGP_TRUST_OBJ_free(ret -> trust_cert);

		if (ret->name != NULL) OPENSSL_free(ret->name);

		break;

	}

	return 1;

}

ASN1_SEQUENCE_ref(XPGP, xPGP_cb, CRYPTO_LOCK_X509) = {
	ASN1_SIMPLE(XPGP, version, ASN1_INTEGER),
	ASN1_SIMPLE(XPGP, key, XPGP_PUBKEY),
	ASN1_SIMPLE(XPGP, subject, XPGP_UID),
	ASN1_SEQUENCE_OF(XPGP, signs, XPGP_SIGNATURE)
} ASN1_SEQUENCE_END_ref(XPGP, XPGP)

IMPLEMENT_ASN1_FUNCTIONS(XPGP)
IMPLEMENT_ASN1_DUP_FUNCTION(XPGP)

static ASN1_METHOD meth={
	(int (*)())  i2d_XPGP,
	(char *(*)())d2i_XPGP,
	(char *(*)())XPGP_new,
	(void (*)()) XPGP_free};

ASN1_METHOD *XPGP_asn1_meth(void)
	{
	return(&meth);
	}
