/* crypto/xpgp/xPGP_aux.c */
/*
 * Copyright (C) 2005-2006 Mark Fernie (drbob@lunamutt.com)  
 * All rights reserved.
 */
/* ====================================================================
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 *
 */


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef OPENSSL_NO_STDIO
#define APPS_WIN16
#endif
#include <openssl/bio.h>
#include <openssl/asn1.h>
#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/pem.h>

#include <openssl/xPGP.h>

#include <openssl/ssl.h>

/****************** Functions to generate a XPGP Certificate ***************
 */

int generate_xpgp(
		const char *pgp_file, const char *pk_file, const char *passwd,
		const char *name, const char *email, const char *org, 
		const char *loc, const char *state, const char *country, int nbits_in); 

void generate_XPGP(XPGP **pgp, EVP_PKEY **pkey,
		const char *name, const char *email, const char *org, 
		const char *loc, const char *state, const char *country, int nbits_in);


void sign_xpgp(const char *subj_file, const char *issuer_file, 
		const char *ipk_file, const char *ipk_passwd);

void sign_XPGP(XPGP *subj, XPGP *issuer, EVP_PKEY *privKey);

XPGP *load_xpgp(const char *file);
EVP_PKEY *load_privKey(const char *pk_file, const char *pk_passwd);

XPGP_KEYRING *xpgp_init(const char *pgp_file, const char *pk_file, const char *pk_passwd);
void display_xpgp(const char *fname);

void display_XPGP(XPGP *pgp, FILE *out);

int xpgp_display_block(const unsigned char *data, unsigned int len, 
	const unsigned char *title);


int generate_xpgp(
		const char *pgp_file, const char *pk_file, const char *passwd,
		const char *name, const char *email, const char *org, 
		const char *loc, const char *state, const char *country, int nbits_in)
{
        // setup output.
	BIO *bio_out = NULL;
	bio_out = BIO_new(BIO_s_file());
	BIO_set_fp(bio_out,stdout,BIO_NOCLOSE);

	EVP_PKEY *pkey = NULL;

	// first generate a key....
	if ((pkey=EVP_PKEY_new()) == NULL)
	{
		fprintf(stderr,"generate_xpgp: Couldn't Create Key\n");
		return 0;
	}

	int nbits = 2048;
	unsigned long e = 0x10001;

	if ((nbits_in >= 512) && (nbits_in <= 4096))
	{
		nbits = nbits_in;
	}
	else
	{
		fprintf(stderr,"generate_xpgp: strange num of nbits: %d\n", nbits_in);
		fprintf(stderr,"generate_xpgp: reverting to %d\n", nbits);
	}


	RSA *rsa = RSA_generate_key(nbits, e, NULL, NULL);
	if ((rsa == NULL) || !EVP_PKEY_assign_RSA(pkey, rsa))
	{
		if(rsa) RSA_free(rsa);
		fprintf(stderr,"generate_xpgp: Couldn't Generate RSA Key!\n");
		return 0;
	}

	// open the file.
	FILE *out;
	if (NULL == (out = fopen(pk_file, "w")))
	{
		fprintf(stderr,"generate_xpgp: Couldn't Create Key File!");
		fprintf(stderr," : %s\n", pk_file);
		return 0;
	}

	const EVP_CIPHER *cipher = EVP_des_ede3_cbc();

	if (!PEM_write_PrivateKey(out,pkey,cipher,
			NULL,0,NULL,(void *) passwd))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Save Private Key");
		fprintf(stderr," : %s\n", pk_file);
		return 0;
	}
	fclose(out);

	// We have now created a private key....
	fprintf(stderr,"generate_xpgp() Saved Private Key");
	fprintf(stderr," : %s\n", pk_file);

	/********** Test Loading the private Key.... ************/
	FILE *tst_in = NULL;
	EVP_PKEY *tst_pkey = NULL;
        if (NULL == (tst_in = fopen(pk_file, "rb")))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Open Private Key");
		fprintf(stderr," : %s\n", pk_file);
		return 0;
	}

	if (NULL == (tst_pkey = 
		PEM_read_PrivateKey(tst_in,NULL,NULL,(void *) passwd)))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Read Private Key");
		fprintf(stderr," : %s\n", pk_file);
		return 0;
	}
	fclose(tst_in);
	EVP_PKEY_free(tst_pkey);
	/********** Test Loading the private Key.... ************/


	// Create a XPGP Certificate.
	XPGP *xpgp = XPGP_new();
	// Are the sub parts auto created?? Should be.
	//XPGP_PUBKEY *xpgp_key = XPGP_PUBKEY_new();
	//XPGP_UID *xpgp_uid = XPGP_UID_new();

	unsigned long chtype = MBSTRING_ASC;

	// fill in the request.

	/**** XPGP -> Version ********************************/
	if (!XPGP_set_version(xpgp,0L)) /* version 1 */
	{
		fprintf(stderr,"generate_xpgp(): Couldn't Set Version!\n");
		return 0;
	}
	/**** XPGP -> Version ********************************/
	/**** XPGP -> UID     ********************************/

	// create the name.
	X509_NAME *xpgp_name = X509_NAME_new();

	// fields to add.
	// commonName CN
	// emailAddress (none)
	// organizationName O
	// localityName L
	// stateOrProvinceName ST
	// countryName C

	if (0 < strlen(name))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "CN", chtype, 
			(unsigned char *) name, -1, -1, 0);
	}
	else
	{
		fprintf(stderr,"generate_xpgp(): No Name -> Not creating XPGP Cert!\n");
		return 0;
	}

	if (0 < strlen(email))
	{
		//X509_NAME_add_entry_by_txt(xpgp_name, "Email", 0, 
		//  (unsigned char *) ui -> gen_email -> value(), -1, -1, 0);
		X509_NAME_add_entry_by_NID(xpgp_name, 48, 0, 
			(unsigned char *) email, -1, -1, 0);
	}

	if (0 < strlen(org))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "O", chtype, 
			(unsigned char *) org, -1, -1, 0);
	}

	if (0 < strlen(loc))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "L", chtype, 
			(unsigned char *) loc, -1, -1, 0);
	}

	if (0 < strlen(state))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "ST", chtype, 
			(unsigned char *) state, -1, -1, 0);
	}

	if (0 < strlen(country))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "C", chtype, 
			(unsigned char *) country, -1, -1, 0);
	}

	if (!XPGP_set_uid_name(xpgp,xpgp_name))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set Name to Request!\n");
		X509_NAME_free(xpgp_name);
		return 0;
	}

	X509_NAME_free(xpgp_name);
	/**** XPGP -> UID     ********************************/

	/**** XPGP -> Key     ********************************/

	if (!XPGP_PUBKEY_set_version(xpgp->key,0L)) /* version 1 */
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set PUBKEY Version!\n");
		return 0;
	}

	ASN1_TIME *start_time = M_ASN1_TIME_new();
	ASN1_TIME *end_time = M_ASN1_TIME_new();

	// set the current time.
	if (!X509_gmtime_adj(start_time, 0))
	{
		fprintf(stderr,"generate_xpgp(): Set Start Time Not Possible!\n");
		return 0;
	}

	// adjust by 5 years.... (3600 * 24 * 365 * 5)
	if (!X509_gmtime_adj(end_time, 3600 * 24 * 365 * 5))
	{
		fprintf(stderr,"generate_xpgp(): Set End Time Not Possible!\n");
		return 0;
	}

	XPGP_PUBKEY_set_notBefore(xpgp->key, start_time);
	XPGP_PUBKEY_set_notAfter( xpgp->key, end_time);

	/* need to clean up */
	M_ASN1_TIME_free(start_time);
	M_ASN1_TIME_free(end_time);

	if (!XPGP_PUBKEY_set_pubkey(xpgp->key,pkey))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set Key to Request!\n");
		return 0;
	}
	/**** XPGP -> Key     ********************************/
	
	/**** XPGP -> SelfSign *******************************/

	const EVP_MD *digest=EVP_sha1();

	if (!XPGP_sign_it(xpgp,xpgp,pkey,digest))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Sign Request!\n");
		return 0;
	}
	fprintf(stderr,"generate_xpgp() Sucessfully Self-Signed Certificate!\n");

	/**** XPGP -> SelfSign *******************************/
	/**** XPGP -> Verify   *******************************/
//#ifdef NOT_DEFINED

	/* Get the key, and the signature */
	XPGP_SIGNATURE *ssign = sk_XPGP_SIGNATURE_value(xpgp->signs, 0);
	EVP_PKEY      *tmpkey = XPGP_get_pubkey(xpgp);

	if (!XPGP_verify_it(xpgp, ssign, tmpkey))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Verify Self-Signature!\n");
		return 0;
	}

	fprintf(stderr,"generate_xpgp(): Verifed Self-Signature!\n");

	EVP_PKEY_free(tmpkey);
//#endif
	/**** XPGP -> Verify   *******************************/

	// open the file.
	if (NULL == (out = fopen(pgp_file, "w")))
	{
		fprintf(stderr,"generate_xpgp() Couldn't create Cert File");
		fprintf(stderr," : %s\n", pgp_file);
		return 0;
	}

	if (!PEM_write_XPGP(out,xpgp))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Save Cert Req");
		fprintf(stderr," : %s\n", pk_file);
		return 0;
	}
	fclose(out);

	fprintf(stderr,"generate_xpgp() Saved Certificate: ");
	fprintf(stderr," : %s\n", pgp_file);
	fprintf(stderr,"Printing Certificate:\n");

	XPGP_print(bio_out, xpgp);

	// clean up....
	XPGP_free(xpgp);

	return 1;
}


void generate_XPGP(XPGP **pgp, EVP_PKEY **pkey,
		const char *name, const char *email, const char *org, 
		const char *loc, const char *state, const char *country, int nbits_in)
{
	EVP_PKEY *xpkey = NULL;
	XPGP *xpgp = NULL;

	// first generate a key....
	if ((xpkey=EVP_PKEY_new()) == NULL)
	{
		fprintf(stderr,"generate_XPGP: Couldn't Create Key\n");
		return;
	}

	int nbits = 2048;
	unsigned long e = 0x10001;

	if ((nbits_in >= 512) && (nbits_in <= 4096))
	{
		nbits = nbits_in;
	}
	else
	{
		fprintf(stderr,"generate_xpgp: strange num of nbits: %d\n", nbits_in);
		fprintf(stderr,"generate_xpgp: reverting to %d\n", nbits);
	}


	RSA *rsa = RSA_generate_key(nbits, e, NULL, NULL);
	if ((rsa == NULL) || !EVP_PKEY_assign_RSA(xpkey, rsa))
	{
		if(rsa) RSA_free(rsa);
		fprintf(stderr,"generate_XPGP: Couldn't Generate RSA Key!\n");
		return;
	}

	// We have now created a private key....
	fprintf(stderr,"generate_XPGP(): Successfully created Private Key:\n");

	/********** Test Loading the private Key.... ************/
	// Create a XPGP Certificate.
	xpgp = XPGP_new();
	// Are the sub parts auto created?? Should be.
	//XPGP_PUBKEY *xpgp_key = XPGP_PUBKEY_new();
	//XPGP_UID *xpgp_uid = XPGP_UID_new();

	unsigned long chtype = MBSTRING_ASC;

	// fill in the request.

	/**** XPGP -> Version ********************************/
	if (!XPGP_set_version(xpgp,0L)) /* version 1 */
	{
		fprintf(stderr,"generate_XPGP() Couldn't Set Version!\n");
		return;
	}
	/**** XPGP -> Version ********************************/
	/**** XPGP -> UID     ********************************/

	// create the name.
	X509_NAME *xpgp_name = X509_NAME_new();

	// fields to add.
	// commonName CN
	// emailAddress (none)
	// organizationName O
	// localityName L
	// stateOrProvinceName ST
	// countryName C

	if (0 < strlen(name))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "CN", chtype, 
			(unsigned char *) name, -1, -1, 0);
	}
	else
	{
		fprintf(stderr,"generate_XPGP() No Name -> Not creating XPGP Cert!\n");
		return;
	}

	if (0 < strlen(email))
	{
		//X509_NAME_add_entry_by_txt(xpgp_name, "Email", 0, 
		//  (unsigned char *) ui -> gen_email -> value(), -1, -1, 0);
		X509_NAME_add_entry_by_NID(xpgp_name, 48, 0, 
			(unsigned char *) email, -1, -1, 0);
	}

	if (0 < strlen(org))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "O", chtype, 
			(unsigned char *) org, -1, -1, 0);
	}

	if (0 < strlen(loc))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "L", chtype, 
			(unsigned char *) loc, -1, -1, 0);
	}

	if (0 < strlen(state))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "ST", chtype, 
			(unsigned char *) state, -1, -1, 0);
	}

	if (0 < strlen(country))
	{
		X509_NAME_add_entry_by_txt(xpgp_name, "C", chtype, 
			(unsigned char *) country, -1, -1, 0);
	}

	if (!XPGP_set_uid_name(xpgp,xpgp_name))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set Name to Request!\n");
		X509_NAME_free(xpgp_name);
		return;
	}

	X509_NAME_free(xpgp_name);
	/**** XPGP -> UID     ********************************/

	/**** XPGP -> Key     ********************************/

	if (!XPGP_PUBKEY_set_version(xpgp->key,0L)) /* version 1 */
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set PUBKEY Version!\n");
		return;
	}

	ASN1_TIME *start_time = M_ASN1_TIME_new();
	ASN1_TIME *end_time = M_ASN1_TIME_new();

	// set the current time.
	if (!X509_gmtime_adj(start_time, 0))
	{
		fprintf(stderr,"generate_xpgp(): Set Start Time Not Possible!\n");
		return;
	}

	// adjust by 5 years.... (3600 * 24 * 365 * 5)
	if (!X509_gmtime_adj(end_time, 3600 * 24 * 365 * 5))
	{
		fprintf(stderr,"generate_xpgp(): Set End Time Not Possible!\n");
		return;
	}

	XPGP_PUBKEY_set_notBefore(xpgp->key, start_time);
	XPGP_PUBKEY_set_notAfter( xpgp->key, end_time);

	/* need to clean up */
	M_ASN1_TIME_free(start_time);
	M_ASN1_TIME_free(end_time);

	if (!XPGP_PUBKEY_set_pubkey(xpgp->key,xpkey))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Set Key to Request!\n");
		return;
	}
	/**** XPGP -> Key     ********************************/
	
	/**** XPGP -> SelfSign *******************************/

	const EVP_MD *digest=EVP_sha1();

	if (!XPGP_sign_it(xpgp,xpgp,xpkey,digest))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Sign Request!\n");
		return;
	}
	fprintf(stderr,"generate_xpgp() Sucessfully Self-Signed Certificate!\n");

	/**** XPGP -> SelfSign *******************************/
	/**** XPGP -> Verify   *******************************/
//#ifdef NOT_DEFINED

	/* Get the key, and the signature */
	XPGP_SIGNATURE *ssign = sk_XPGP_SIGNATURE_value(xpgp->signs, 0);
	EVP_PKEY      *tmpkey = XPGP_get_pubkey(xpgp);

	if (!XPGP_verify_it(xpgp, ssign, tmpkey))
	{
		fprintf(stderr,"generate_xpgp() Couldn't Verify Self-Signature!\n");
		return;
	}
	fprintf(stderr,"generate_xpgp(): Verifed Self-Signature!\n");

	EVP_PKEY_free(tmpkey);
//#endif
	/**** XPGP -> Verify   *******************************/

	*pgp = xpgp;
	*pkey = xpkey;

	return;

}


/****************** Functions to display a XPGP Certificate ***************
 */


void sign_xpgp(const char *subj_file, const char *signer_file, 
		const char *pk_file, const char *pk_passwd)
{
	FILE *in, *out;
	XPGP *xpgp_subj = NULL;
	XPGP *xpgp_signer = NULL;
	EVP_PKEY *pkey = NULL;

	/************ Setup the BIO output ********************/
	// setup stdout.
	BIO *bio_out = NULL;
	bio_out = BIO_new(BIO_s_file());
	BIO_set_fp(bio_out,stdout,BIO_NOCLOSE);

	/************ Load the subj **************************/
	if (NULL == (in = fopen(subj_file, "r")))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Open xpgp File!");
		fprintf(stderr," : %s\n", subj_file);
		return;
	}

	if (!PEM_read_XPGP(in,&xpgp_subj, NULL, NULL))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Read XPGP");
		fprintf(stderr," : %s\n", subj_file);
		return;
	}
	fclose(in);

	/************ Load the issuer (us) *******************/
	if (NULL == (in = fopen(signer_file, "r")))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Open XPGP Sig Issuer");
		fprintf(stderr," : %s\n", signer_file);
		return;
	}

	if (!PEM_read_XPGP(in,&xpgp_signer, NULL, NULL))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Read XPGP Sig Issuer");
		fprintf(stderr," : %s\n", signer_file);
		return;
	}
	fclose(in);

	/************ Load the private key *******************/
	// open the file.
        if (NULL == (in = fopen(pk_file, "rb")))
        {
		fprintf(stderr,"sign_xpgp() Couldn't Open PrivKey File");
		fprintf(stderr," : %s\n", pk_file);
		return;
	}

        if (NULL == (pkey = 
		//PEM_read_PrivateKey(in,NULL,NULL,(void *) pk_passwd)))
		PEM_read_PrivateKey(in,NULL,NULL,NULL)))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Read PrivKey File");
		fprintf(stderr," : %s\n", pk_file);
		return;
	}
        fclose(in);

	/************ Sign the XPGP Cert   *******************/

	const EVP_MD *digest=EVP_sha1();

	if (!XPGP_sign_it(xpgp_subj,xpgp_signer,pkey,digest))
	{
		fprintf(stderr,"sign_xpgp() Couldn't SIgn Request!\n");
		return;
	}

	/************ Print New Cert       *******************/
	XPGP_print(bio_out, xpgp_subj);
	//return;

	/************ Save the New Cert    *******************/
	// open the file.
	if (NULL == (out = fopen(subj_file, "w")))
	{
		fprintf(stderr,"sign_xpgp() Couldn't Create Request File\n");
		fprintf(stderr," : %s\n", subj_file);
		return;
	}

	if (!PEM_write_XPGP(out,xpgp_subj))
	{
		fprintf(stderr,"sign_xpgp(): Couldn't Save Certificate Request!\n");
		return;
	}
	fclose(out);

	XPGP_free(xpgp_subj);
	XPGP_free(xpgp_signer);
	EVP_PKEY_free(pkey);
	return;
}



void sign_XPGP(XPGP *subj, XPGP *issuer, 
		EVP_PKEY *privKey)
{
	const EVP_MD *digest=EVP_sha1();

	if (!XPGP_sign_it(subj,issuer,privKey,digest))
	{
		fprintf(stderr,"sign_XPGP: Couldn't Sign Request!\n");
		return;
	}

	return;
}




XPGP *load_xpgp(const char *file)
{
	FILE *in;
	XPGP *xpgp = NULL;
	/************ Load the subj **************************/
	if (NULL == (in = fopen(file, "r")))
	{
		fprintf(stderr,"load_xpgp() Couldn't Open xpgp file: %s\n",file);
		return NULL;
	}

	if (!PEM_read_XPGP(in,&xpgp, NULL, NULL))
	{
		fprintf(stderr,"load_xpgp() Couldn't Read XPGP: %s\n",file);
		fclose(in);
		return NULL;
	}
	fclose(in);

	/************ Load the issuer (us) *******************/
	return xpgp;
}



EVP_PKEY *load_privKey(const char *pk_file, const char *pk_passwd)
{
	FILE *in;
	EVP_PKEY *pkey = NULL;

	/************ Load the private key *******************/
	// open the file.
        if (NULL == (in = fopen(pk_file, "rb")))
        {
		fprintf(stderr,"load_privKey() Couldn't Open Private Key File!");
		fprintf(stderr,"%s\n", pk_file);
		return NULL;
	}

	if (pk_passwd)
	{
        	pkey = PEM_read_PrivateKey(in,NULL,NULL,(void *) pk_passwd);
	}
	else
	{
		pkey = PEM_read_PrivateKey(in,NULL,NULL,NULL);
	}

        if (pkey == NULL)
	{
		fprintf(stderr,"load_privKey() Couldn't Read Private Key!\n");
        	fclose(in);
		return NULL;
	}
        fclose(in);

	return pkey;
}


XPGP_KEYRING *xpgp_init(const char *pgp_file, const char *pk_file, const char *pk_passwd)
{
	FILE *in;
	XPGP *xpgp = NULL;
	EVP_PKEY *pkey = NULL;
	XPGP_KEYRING *kr = NULL;

	/************ Setup the BIO output ********************/
	// setup stdout.
	BIO *bio_out = NULL;
	bio_out = BIO_new(BIO_s_file());
	BIO_set_fp(bio_out,stdout,BIO_NOCLOSE);

	/************ Load the subj **************************/
	if (NULL == (in = fopen(pgp_file, "r")))
	{
		fprintf(stderr, "xpgp_init() Couldn't Open %s\n", pgp_file);
		return NULL;
	}

	if (!PEM_read_XPGP(in,&xpgp, NULL, NULL))
	{
		fprintf(stderr, "xpgp_init() Couldn't Read XPGP %s\n", pgp_file);
		return NULL;
	}
	fclose(in);

	/************ Load the private key *******************/
	// open the file.
        if (NULL == (in = fopen(pk_file, "rb")))
        {
		fprintf(stderr, "xpgp_init() Couldn't Open %s\n", pk_file);
		return NULL;
	}

        if (NULL == (pkey = 
		PEM_read_PrivateKey(in,NULL,NULL,(void *) pk_passwd)))
		//PEM_read_PrivateKey(in,NULL,NULL,NULL)))
	{
		fprintf(stderr, "xpgp_init() Couldn't Read PrivKey %s\n", pk_file);
		return NULL;
	}
        fclose(in);

	kr = createPGPContext(xpgp, pkey);

	return kr;
}


void display_xpgp(const char *fname)
{
	// open the file.
	FILE *in;
	if (NULL == (in = fopen(fname, "r")))
	{
		fprintf(stderr, "display_xpgp() Couldn't Open %s\n", fname);
		return;
	}

	XPGP *xpgp = NULL;
	if (!PEM_read_XPGP(in,&xpgp, NULL, NULL))
	{
		fprintf(stderr, "display_xpgp() Couldn't Read XPGP %s\n", fname);
		return;
	}
	fclose(in);

	display_XPGP(xpgp, stdout);

	XPGP_free(xpgp);
	return;
}

void display_XPGP(XPGP *xpgp, FILE *toWhere)
{
	// setup output.
	BIO *out = NULL;
	out = BIO_new(BIO_s_file());
	BIO_set_fp(out,toWhere,BIO_NOCLOSE);

	XPGP_print(out, xpgp);
	return;
}



int xpgp_display_block(const unsigned char *data, const unsigned int len, 
	const unsigned char *title)
{
	unsigned int i = 0;
	fprintf(stderr, "%s(%d) :\n", title, len);
	for(i = 0; i < len; i++)
	{
		fprintf(stderr, "%02x", data[i]);
		if ((i+1)%16==0)
		{
			fprintf(stderr, "\n");
		}
		else
		{
			fprintf(stderr, ":");
		}
	}
	fprintf(stderr, "\n");
	return 1;
}

			



