/* crypto/xpgp/xPGP_sign.c */
/*
 * Copyright (C) 2005-2006 Mark Fernie (drbob@lunamutt.com)  
 * All rights reserved.
 */
/* ====================================================================
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 *
 */

#include <stdio.h>
#include "cryptlib.h"
#include <openssl/asn1.h>
#include <openssl/objects.h>
#include <openssl/evp.h>
#include <openssl/xPGP.h>


/* High Level Verify Function */
int XPGP_verify_it(XPGP *subj, XPGP_SIGNATURE *sign, EVP_PKEY *pkey)
{
	int ret = 0;

#ifdef XPGP_DEBUG
	fprintf(stderr, "XPGP_verify_it()\n");
	fprintf(stderr, "XPGP_verify_it() : XPGP subj *******************\n");
	display_XPGP(subj, stderr);
	fprintf(stderr, "XPGP_verify_it() : *****************************\n");
#endif /* XPGP_DEBUG */

	/* create data types */
	XPGP_SIGNDATA *sd = XPGP_SIGNDATA_new();

	if (!XPGP_setup_data_verify(sd, subj, sign))
	{
	   fprintf(stderr, "XPGP_verify_it: XPGP_setup_data_verify failed\n");
	   goto err;
	}

	ret = XPGP_verify(sign, sd, pkey);


err:
	XPGP_SIGNDATA_free(sd);
	fprintf(stderr, "XPGP_verify_it() returns: %d\n", ret);
	if (!ret)
	{
		/* Errors in verify */
        	fprintf(stderr, "XPGP_verify_it() any errors?\n");
        	ERR_print_errors_fp(stderr);
		ERR_clear_error();
	}
	return ret;
}

/* High Level Sign Function */
int XPGP_sign_it(XPGP *subj, XPGP *issuer, EVP_PKEY *pkey, const EVP_MD *md)
{
	/* create data types */
	XPGP_SIGNDATA *sd = XPGP_SIGNDATA_new();
	XPGP_SIGNATURE *sign = XPGP_SIGNATURE_new();

	fprintf(stderr, "XPGP_sign_it()\n");

	if (!XPGP_setup_data_sign(sd, subj, issuer))
	{
		fprintf(stderr, "XPGP_sign_it: XPGP_setup_signdata failed\n");
		goto err;
	}

	if (!XPGP_setup_signature(sign, sd, issuer))
	{
		fprintf(stderr, "XPGP_sign_it: XPGP_setup_signature failed\n");
		goto err;
	}

	if (!XPGP_sign(sign, sd, pkey, md))
	{
		fprintf(stderr, "XPGP_sign_it: XPGP_sign failed\n");
		goto err;
	}

	if (!XPGP_add_signature(subj, sign))
	{
		fprintf(stderr, "XPGP_sign_it: XPGP_add_signature failed\n");
		goto err;
	}


	XPGP_SIGNDATA_free(sd);
	return 1;
err:
        fprintf(stderr, "XPGP_sign_it() any errors?\n");
        ERR_print_errors_fp(stderr);
	ERR_clear_error();

	XPGP_SIGNDATA_free(sd);
	XPGP_SIGNATURE_free(sign);
	return 0;
}

/* Lower Helper Fns */
int XPGP_setup_data_sign(XPGP_SIGNDATA *d, XPGP *s, XPGP *i)
{
	if ((d == NULL) || (s == NULL) || (i == NULL))
		return (0);

	// set version.
	if (d->version == NULL)
		{
		if ((d->version=M_ASN1_INTEGER_new()) == NULL)
			return(0);
		}

	if (!ASN1_INTEGER_set(d->version, 0L))
		return (0);

	// set type.
	if (d->type == NULL)
		{
		if ((d->type=M_ASN1_INTEGER_new()) == NULL)
			return(0);
		}

	if (!ASN1_INTEGER_set(d->type, 1L))
		return (0);

	// timestamp
	if (d->timestamp == NULL)
		{
		if ((d->timestamp = M_ASN1_TIME_new()) == NULL)
			return (0);
		}

	if (!X509_gmtime_adj(d->timestamp, 0))
		return (0);

	// set validity.
	if (s->key == NULL)
		return (0);

	if (s->key->validity == NULL)
		return (0);

	if (!d->validity)
		X509_VAL_free(d->validity);

	if (!(d->validity = X509_VAL_new()))
		return (0);
	M_ASN1_TIME_free(d->validity->notBefore);
	M_ASN1_TIME_free(d->validity->notAfter);
	d->validity->notBefore = M_ASN1_TIME_dup(s->key->validity->notBefore);
	d->validity->notAfter  = M_ASN1_TIME_dup(s->key->validity->notAfter);

	// key
	if ((s->key == NULL) || (s->key->key == NULL))
		return (0);

	// bug here - why would freeing it cause a segfault.
	if (d->key)
	{
		X509_PUBKEY_free(d->key);
		d->key = NULL;
	}

	EVP_PKEY *tmp_pkey = X509_PUBKEY_get(s->key->key);
	if(!X509_PUBKEY_set(&d->key, tmp_pkey))
	{
		EVP_PKEY_free(tmp_pkey);
		return (0);
	}
	// must freed
	EVP_PKEY_free(tmp_pkey);

	// subject name.
	if (!X509_NAME_set(&d->subject, s->subject->subject))
		return (0);

	// issuer name.
	if (!X509_NAME_set(&d->issuer,  i->subject->subject))
		return (0);

	// signtype ... filled at signing.
	return 1;
}

int XPGP_setup_data_verify(XPGP_SIGNDATA *d, XPGP *x, XPGP_SIGNATURE *s)
{
	if ((d == NULL) || (x == NULL) || (s == NULL))
		return (0);

	// set version. (from signature)
	if (!s->version) return 0;
	if (!d->version)
		{
		if (!(d->version=M_ASN1_INTEGER_new()))
			return(0);
		}

	if (!ASN1_INTEGER_set(d->version, ASN1_INTEGER_get(s->version)))
		return (0);

	// set type. (from signature)
	if (!s->type) return 0;
	if (!d->type)
		{
		if (!(d->type=M_ASN1_INTEGER_new()))
			return(0);
		}

	if (!ASN1_INTEGER_set(d->type, ASN1_INTEGER_get(s->type)))
		return (0);

	// timestamp (from signature)
	if (!s->timestamp) return 0;
	if (d->timestamp) M_ASN1_TIME_free(d->timestamp);
	if (!(d->timestamp = M_ASN1_TIME_dup(s->timestamp))) return (0);

	// set validity. (from xpgp)
	if ((!x->key) || (!x->key->validity)) return (0);

	if (!d->validity)
		{
		if (!(d->validity= X509_VAL_new()))
			return(0);
		}

	M_ASN1_TIME_free(d->validity->notBefore);
	M_ASN1_TIME_free(d->validity->notAfter);
	d->validity->notBefore = M_ASN1_TIME_dup(x->key->validity->notBefore);
	d->validity->notAfter  = M_ASN1_TIME_dup(x->key->validity->notAfter);

	// set key (from pgp)
	if ((!x->key) || (!x->key->key))
		return (0);

	if (d->key)
	{
		X509_PUBKEY_free(d->key);
		d->key=NULL;
	}

	EVP_PKEY *tmp_pkey = X509_PUBKEY_get(x->key->key);
	if(!X509_PUBKEY_set(&d->key, tmp_pkey))
	{
		EVP_PKEY_free(tmp_pkey);
		return (0);
	}
	// must freed
	EVP_PKEY_free(tmp_pkey);

	// subject name. (from xpgp)
	if (!X509_NAME_set(&d->subject, x->subject->subject))
		return (0);

	// issuer name. (from signature)
	if (!X509_NAME_set(&d->issuer,  s->issuer))
		return (0);

	// signtype (from signature)
	if (!s->signtype) return (0);
	if (d->signtype)
		X509_ALGOR_free(d->signtype);
	if (!(d->signtype=X509_ALGOR_dup(s->signtype)))
		return (0);
	return 1;
}

/* issuer is only needed for the keyid.... */
int XPGP_setup_signature(XPGP_SIGNATURE *s, XPGP_SIGNDATA *d, XPGP *i)
{
	/* transfer data to the new structure */

	if ((d == NULL) || (s == NULL) || (i == NULL))
		return (0);

	// set version.
	if (d->version == NULL)
		return (0);

	if (!s->version)
	{
		M_ASN1_INTEGER_free(s->version);
	}
	if ((s->version=ASN1_INTEGER_dup(d->version)) == NULL)
		return(0);

	// set type.
	if (d->type == NULL)
		return (0);

	if (!s->type)
	{
		M_ASN1_INTEGER_free(s->type);
	}
	if ((s->type=ASN1_INTEGER_dup(d->type)) == NULL)
		return(0);

	// set timestamp.
	if (d->timestamp == NULL)
		return (0);

	if (!s->timestamp)
		M_ASN1_TIME_free(s->timestamp);

	if (!(s->timestamp = M_ASN1_TIME_dup(d->timestamp)))
		return (0);

	// issuer name.
	if (!X509_NAME_set(&s->issuer,  d->issuer))
		return (0);

	// deal with keyid in a bit.
	//if (!s -> keyid)
	//s -> keyid = XPGP_calc_keyid(issuer);
	// key

	// signtype  ... filled at signing.
	// signature ... filled at signature.

	return 1;
}

int XPGP_add_signature(XPGP *x, XPGP_SIGNATURE *s)
{
	/* transfer data to the new structure */
	if ((!s) || (!x))
		return 0;

        if(!x->signs && !(x->signs = sk_XPGP_SIGNATURE_new_null())) return 0;

	return (sk_XPGP_SIGNATURE_push(x->signs, s));
}


