/* crypto/xpgp/xPGP_vfy.c */
/*
 * Copyright (C) 2005-2006 Mark Fernie (drbob@lunamutt.com)  
 * All rights reserved.
 */
/* ====================================================================
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 *
 */

#include <stdio.h>
#include <time.h>
#include <errno.h>

#include "cryptlib.h"
#include <openssl/crypto.h>
#include <openssl/lhash.h>
#include <openssl/buffer.h>
#include <openssl/evp.h>
#include <openssl/asn1.h>
#include <openssl/objects.h>

#include <openssl/xPGP.h>
#include "xPGP_vfy.h"

#define XPGP_KEYRING_STATE_BAD 		-1
#define XPGP_KEYRING_STATE_CHKUNKNOWN  	0
#define XPGP_KEYRING_STATE_OKAY		1


	/* so trust is derived by
	 * 1) If we have signed it.  
	 * 2) The number of Trusted_Signers 
	 * 3) The number of trusted_others 
	 *
	 * If we have signed it -> Trusted.
	 *
	 * Score 0 -> 100;
	 * 0 --- Unknown.
	 * 40 --- TRUST_BASIC  (if one trusted_signer + )
	 * 80 --- TRUST_AUTHEN
	 * 100 --- 
	 *
	 * + (5 - -- per signature of others (
	 *
	 * +(50-100) for own signature.
	 * +(15-40) between 1, 3 requred for BASIC TRUST
	 * +(1-10) (MAX of 40)
	 * - 50
	 */

#define TRUST_BASIC 		40
#define TRUST_AUTHEN		80

#define TRUST_SIGN_OWN_MIN	40
#define TRUST_SIGN_TRSTED_MIN	15
#define TRUST_SIGN_AUTHEN_MIN	1
#define TRUST_SIGN_BASIC_MIN	0

#define TRUST_SIGN_OWN_MAX	80
#define TRUST_SIGN_TRSTED_MAX	40
#define TRUST_SIGN_AUTHEN_MAX	10
#define TRUST_SIGN_BASIC_MAX	5

/* Some new functions to operate on it */
int printX509NameString(FILE *out, X509_NAME *name);
int checkAndMoveSignedByOwnKey(XPGP_KEYRING *kr, STACK_OF(XPGP) *sk);
int 	DisplayStacks(XPGP_KEYRING *kr);
int 	DisplayStack(STACK_OF(XPGP) *sk);
 /**********************************************************************
  * High Level KeyRing Control Fns.
  **********************************************************************
  *
  * int     setHouseKey(XPGP_KEYRING *kr, XPGP *own, EVP_PKEY *privkey)
  * int     setTrustedSigner(XPGP_KEYRING *kr, XPGP *trusted)
  * int     addCertificate(XPGP_KEYRING *kr, XPGP *unknown)
  * int     removeCertificate(XPGP_KEYRING *, XPGP *unwanted)
  * int     checkKeyRing(XPGP_KEYRING *);
  * int     updateUntrustedStacks(XPGP_KEYRING *)
  *
  * int freePGPContext(XPGP_KEYRING *kr);
  * XPGP_KEYRING *createPGPContext(XPGP *own, EVP_PKEY *privkey);
  * 
  */

/* Validation only -> does not transfer signs or certs to keyring 
 */

int	XPGP_sign_certificate(XPGP_KEYRING *kr, XPGP *subj, XPGP *issuer)
{
	return signCertificate(kr, subj, issuer);
}


int XPGP_add_certificate(XPGP_KEYRING *kr, XPGP *x)
{
	return addCertificate(kr, x);
}


int XPGP_remove_certificate(XPGP_KEYRING *kr, XPGP *x)
{
	int ret = removeCertificate(kr, x);
	if (ret)
	{
		checkKeyRing(kr);
	}
	return ret;
}


/* called by ssl_verify_pgp_cert 0 - fail, 1 - pass */

int 	XPGP_verify_cert(XPGP_KEYRING *kr, XPGP *x)
{
	/* if we are not sorted - do so */
	if (kr->state != XPGP_KEYRING_STATE_OKAY)
	{
		checkKeyRing(kr);
	}

	if (kr->superNodeMode)
	{
		/* dangerous mode - but better than
		 * modding the ssl code? 
		 */
		/* this is only check.... */
		if (!XPGP_check_valid_certificate(x))
		{
			return 0;
		}
		return 1;
	}

	if (XPGP_auth_certificate(kr, x) >= TRUST_SIGN_AUTHEN)
	{
		return 1;
	}
	return 0;
}

int XPGP_auth_certificate(XPGP_KEYRING *kr, XPGP *x)
{
	/* first check the certificate */
	if (!XPGP_check_valid_certificate(x))
	{
		return -1;
	}

	return validateCertificate(kr, x);
}



int XPGP_check_valid_certificate(XPGP *x)
{
	/* if first signature is self and correct, 
	 * if dates are okay */
	/* check dates of all signatures? */

	return validateSelfSignedCertificate(x);

	//return 0;
}

/* transfers the new and exciting signatures to 
 * our certificate maximum of 100?
 */

int XPGP_copy_known_signatures(XPGP_KEYRING *kr, XPGP *dest, XPGP *src)
{
	/* only copy signatures that we know about, 
	 * don't want to propagate crap
	 */
	int cpyd = 0;
	int i,j;
	STACK_OF(XPGP_SIGNATURE) *signstore = sk_XPGP_SIGNATURE_new_null();

	/* if dest has more than max, forget it */

	/* iterate through src signs */
	XPGP_SIGNATURE *sign;
	for(i = 1; i < sk_XPGP_SIGNATURE_num(src->signs); i++)
	{
		sign = sk_XPGP_SIGNATURE_value(src->signs, i);

		fprintf(stderr, "Copy_Signatures: Checking for: ");
		printX509NameString(stderr, 
				XPGP_SIGNATURE_get_issuer_name(sign));
		fprintf(stderr, "\n");
		/* for each, check if it exists in the dest */
		/* if not
		 * 	check the TrustLevel of the signature
		 * 	if Auth or above, then keep signature.
		 */
        	XPGP_SIGNATURE *sign2;

		int found = 0;
        	for (j = 1; (!found) && (j < sk_XPGP_SIGNATURE_num(dest->signs)); j++)
        	{
               		sign2 = sk_XPGP_SIGNATURE_value(dest->signs, j);
			/* should now work! */
			if (0 == XPGP_SIGNATURE_cmp(sign,sign2))
			{
				fprintf(stderr, "Copy_Signatures: \tFound Identical for: ");
				printX509NameString(stderr, 
					XPGP_SIGNATURE_get_issuer_name(sign));
				fprintf(stderr, "\n");

				found = 1;
			}
		}

		if (found)
		{
			continue;
		}

		/* check for any value */
		fprintf(stderr, "Copy_Signatures: \tChecking if Signature is useful ");
		printX509NameString(stderr, 
			XPGP_SIGNATURE_get_issuer_name(sign));
		fprintf(stderr, "\n");
	
		int ret;
		if (TRUST_SIGN_AUTHEN <= (ret = getTrustSignatureLevel(kr, src, sign)))
		{
			/* add to our stack */
               		sk_XPGP_push(signstore, sign);
                }
		fprintf(stderr, "Copy_Signatures: \tTrustLevelReturned: %d\n", ret);
        }

	/* copy any on tmp stack to the end of dest stack */
        fprintf(stderr,"moveStackToStack()\n");
        for(cpyd = 0; 0 < sk_XPGP_SIGNATURE_num(signstore); cpyd++)
        {
               sign = sk_XPGP_SIGNATURE_pop(signstore);
	       /* now make a copy 
		* or at least add a reference 
		*/
		fprintf(stderr, "Copying Signature over: ");
		printX509NameString(stderr, XPGP_SIGNATURE_get_issuer_name(sign));
		fprintf(stderr, "\n");

		fprintf(stderr, "Sign Ref: %d Pre-Push!\n", sign->references);

		CRYPTO_add(&(sign->references),1,CRYPTO_LOCK_X509);
                sk_XPGP_SIGNATURE_push(dest->signs, sign);

		fprintf(stderr, "Sign Ref: %d Post-Push!\n", sign->references);
        }
	sk_XPGP_SIGNATURE_free(signstore);
	fprintf(stderr, "%d Signatures Copied\n", cpyd);

	return cpyd;
}

XPGP *XPGP_KEYRING_find_known_cert(XPGP_KEYRING *kr, XPGP *tofind)
{
	/* look through all the certs */
	XPGP *candidate = NULL;
	int i;

	for(i = 0; i < sk_XPGP_num(kr->trusted_ownsign); i++)
	{
		candidate = sk_XPGP_value(kr->trusted_ownsign, i);
		if (0 == XPGP_cmp(tofind, candidate))
		{
			return candidate;
		}
	}

	for(i = 0; i < sk_XPGP_num(kr->trusted_signers); i++)
	{
		candidate = sk_XPGP_value(kr->trusted_signers, i);
		if (0 == XPGP_cmp(tofind, candidate))
		{
			return candidate;
		}
	}

	for(i = 0; i < sk_XPGP_num(kr->trusted_certs); i++)
	{
		candidate = sk_XPGP_value(kr->trusted_certs, i);
		if (0 == XPGP_cmp(tofind, candidate))
		{
			return candidate;
		}
	}

	for(i = 0; i < sk_XPGP_num(kr->trusted_basic); i++)
	{
		candidate = sk_XPGP_value(kr->trusted_basic, i);
		if (0 == XPGP_cmp(tofind, candidate))
		{
			return candidate;
		}
	}

	/* if it is at a level below - not much can help it! */
	return NULL;
}

int XPGP_supernode(XPGP_KEYRING *kr)
{
	return kr -> superNodeMode = 1;
}

XPGP_KEYRING *createPGPContext(XPGP *own, EVP_PKEY *privkey)
{
	XPGP_KEYRING *kr = (XPGP_KEYRING *)OPENSSL_malloc(sizeof(XPGP_KEYRING));

#ifdef XPGP_DEBUG
	fprintf(stderr, "createPGPContext()\n");
#endif /* XPGP_DEBUG */

	kr -> state = XPGP_KEYRING_STATE_BAD;
	kr -> secLevel = 1;
	kr -> error = 0;
	kr -> superNodeMode = 0;
	kr -> trusted_signers = sk_XPGP_new_null();
	kr -> trusted_ownsign = sk_XPGP_new_null();
	kr -> trusted_certs   = sk_XPGP_new_null();
	kr -> trusted_basic   = sk_XPGP_new_null();
	kr -> untrusted_certs = sk_XPGP_new_null();
	kr -> unknown_certs   = sk_XPGP_new_null();

	kr->houseKeyPublic = NULL;
	kr->houseKeyPrivate = NULL;

	setHouseKey(kr, own, privkey);

	return kr;

}


int freePGPContext(XPGP_KEYRING *kr)
{

#ifdef XPGP_DEBUG
	fprintf(stderr, "freePGPContext()\n");
#endif /* XPGP_DEBUG */

	kr -> state = XPGP_KEYRING_STATE_BAD;
	kr -> secLevel = 1;
	kr -> error = 0;
	if (kr -> trusted_signers)
		sk_XPGP_pop_free(kr -> trusted_signers, XPGP_free);
	if (kr -> trusted_ownsign)
		sk_XPGP_pop_free(kr -> trusted_ownsign, XPGP_free);
	if (kr -> trusted_certs)
		sk_XPGP_pop_free(kr -> trusted_certs, XPGP_free);
	if (kr -> trusted_basic)
		sk_XPGP_pop_free(kr -> trusted_basic, XPGP_free);
	if (kr -> untrusted_certs)
		sk_XPGP_pop_free(kr -> untrusted_certs, XPGP_free);
	if (kr -> unknown_certs)
		sk_XPGP_pop_free(kr -> unknown_certs, XPGP_free);

	if (kr->houseKeyPublic)
		XPGP_free(kr->houseKeyPublic);
	if (kr->houseKeyPrivate)
		EVP_PKEY_free(kr->houseKeyPrivate);

	/* free the mem finally */
	OPENSSL_free(kr);

	return (1);
}

int     setHouseKey(XPGP_KEYRING *kr, XPGP *own, EVP_PKEY *privkey)
{
#ifdef XPGP_DEBUG
	fprintf(stderr, "setHouseKey()\n");
#endif /* XPGP_DEBUG */

	// save and add references to keys.
	kr->houseKeyPublic = own;
	CRYPTO_add(&(kr->houseKeyPublic->references),1,CRYPTO_LOCK_X509);

	kr->houseKeyPrivate = privkey;
	CRYPTO_add(&(kr->houseKeyPrivate->references),1,CRYPTO_LOCK_EVP_PKEY);

	kr->state = XPGP_KEYRING_STATE_BAD;

	/*
	return checkKeyRing(kr);
	*/
	return 1;
}

int     XPGP_signer_trusted(XPGP_KEYRING *kr, XPGP *trusted)
{
	int ret = 0;

	fprintf(stderr, "setTrustedSigner()\n");

	/* check that we have signed it */
	ret = checkSignature(trusted, kr->houseKeyPublic);
	if (ret < 0)
	{
		/* not signed by us! */
		fprintf(stderr, "setTrustedSigner() Cert is not signed by us!\n");
		return ret;
	}
	/* is signed */
	/* remove from all stacks */
	removeCertificate(kr, trusted);
	addCertificateToStack(kr->trusted_signers, trusted);

	/* could check keyring */
	return 1;
}


int     XPGP_signer_untrusted(XPGP_KEYRING *kr, XPGP *untrusted)
{
	int ret = 0;

	fprintf(stderr, "XPGP_signer_untrusted()\n");
	if (!removeCertificateFromStack(kr->trusted_signers, untrusted))
	{
		fprintf(stderr, "XPGP_signer_untrusted() Not in Trusted stack\n");
		return 0;
	}

	/* check that we have signed it */
	ret = checkSignature(untrusted, kr->houseKeyPublic);
	if (ret < 0)
	{
		/* not signed by us! */
		addCertificateToStack(kr->unknown_certs, untrusted);
		fprintf(stderr, "setUntrustedSigner() Cert is not signed by us!\n");
		return ret;
	}
	/* is signed */
	addCertificateToStack(kr->trusted_ownsign, untrusted);
	/* need to check the keyring now */

	checkKeyRing(kr);
	return 1;
}


int     addCertificate(XPGP_KEYRING *kr, XPGP *unknown)
{
	fprintf(stderr, "addCertificate()\n");

	/* check that it doesnt exist already */
	if (existsCertificate(kr, unknown))
	{
		/* already in keyring! */
		fprintf(stderr, "addCertificate() Cert is already in keyring!\n");
		return 0;
	}
	/* else add to unknown certificates */
	addCertificateToStack(kr->unknown_certs, unknown);

	/* change ring state */
	if (kr->state > XPGP_KEYRING_STATE_CHKUNKNOWN)
	{
		kr->state = XPGP_KEYRING_STATE_CHKUNKNOWN;
	}

	checkKeyRing(kr);
	return 1;
}


int     removeCertificate(XPGP_KEYRING *kr, XPGP *unwanted)
{
	/* check all stacks */
	int ret = 0;

	fprintf(stderr, "removeCertificate()\n");

	ret |= removeCertificateFromStack(kr->unknown_certs, unwanted);
	ret |= removeCertificateFromStack(kr->untrusted_certs, unwanted);
	ret |= removeCertificateFromStack(kr->trusted_basic, unwanted);
	ret |= removeCertificateFromStack(kr->trusted_certs, unwanted);
	ret |= removeCertificateFromStack(kr->trusted_ownsign, unwanted);
	ret |= removeCertificateFromStack(kr->trusted_signers, unwanted);

	// this is kinda internal fn - don't want to checkKeyRing()
	// each time....
	//checkKeyRing(kr);
	return ret;
}


int     checkKeyRing(XPGP_KEYRING *kr)
{
	int ret;

	fprintf(stderr, "checkKeyRing()\n");

	kr->state = XPGP_KEYRING_STATE_BAD;

	// move all the normal certificates into unknown stack.
	moveStackToStack(kr->unknown_certs, kr->trusted_certs);
	moveStackToStack(kr->unknown_certs, kr->trusted_basic);
	moveStackToStack(kr->unknown_certs, kr->untrusted_certs);
	
	// check that all signatures self-validate.
	validateOrClearCertifateStack(kr->trusted_signers);
	validateOrClearCertifateStack(kr->trusted_ownsign);
	validateOrClearCertifateStack(kr->unknown_certs);

	 // are trusted_signers are signed by us?
	ret = checkOrClearStackSignedByKey(kr->trusted_signers, kr->houseKeyPublic);
	if (!ret)
	{
		/* bad -> weed out bad one!
		 */
		/* already in keyring! */
		fprintf(stderr, "checkKeyRing() Bad Certificate !!! Weed Out\n");
		return ret;
	}
	// are trusted_ownsign are signed by us?
	ret = checkOrClearStackSignedByKey(kr->trusted_ownsign, kr->houseKeyPublic);
	if (!ret)
	{
		/* bad -> weed out bad one!
		 */
		/* already in keyring! */
		fprintf(stderr, "checkKeyRing() Bad Certificate !!! Weed Out Too\n");
		return ret;
	}

	// check if an

	// iterate through trust tree
	kr->state = XPGP_KEYRING_STATE_CHKUNKNOWN;
	return updateUntrustedStacks(kr);
}

int 	updateUntrustedStacks(XPGP_KEYRING *kr)
{
	fprintf(stderr, "updateUntrustedStacks()\n");
	if (kr->state == XPGP_KEYRING_STATE_OKAY)
	{
		return 0;
	}
	int modified = 1;
	int loop = 0;

	checkAndMoveSignedByOwnKey(kr, kr->unknown_certs);

	while((modified != 0) && (++loop != 10))
	{
		modified = 0;
		modified += checkStackTrustworthiness(kr, kr->unknown_certs, 
						TRUST_SIGN_UNKNOWN);
		modified += checkStackTrustworthiness(kr, kr->untrusted_certs, 
						TRUST_SIGN_UNTRUSTED);
		modified += checkStackTrustworthiness(kr, kr->trusted_basic, 
						TRUST_SIGN_BASIC);
	}
	kr->state = XPGP_KEYRING_STATE_OKAY;
	if (loop == 10)
	{
		fprintf(stderr, "Continual Loop!\n");
		kr->state = XPGP_KEYRING_STATE_BAD;
	}
	return 1;
}




 /**********************************************************************
  * Signature related Fns.
  **********************************************************************
  *
  * int	checkSignature(XPGP *candidate, XPGP *validator);
  * int	validateSignature(XPGP *candidate, XPGP_SIGNATURE *signature, XPGP *validator);
  * int findSignature(XPGP *subj, XPGP *issuer, XPGP_SIGNATURE **sign);
  * int	signCertificate(XPGP_KEYRING *kr, XPGP *subj, XPGP *issuer)
  * 
  */




int	checkSignature(XPGP *candidate, XPGP *validator)
{
	/* search through the signature for validator */
	XPGP_SIGNATURE *sign = NULL;
	int ret = 0;
	fprintf(stderr,"checkSignature()\n");


	ret = findSignature(candidate, validator, &sign);
	
	/* if exist - return validateSignature */
	if (ret >= 0)
	{
		return validateSignature(candidate, sign, validator);
	}
	// SHOULD REMOVE CRYPTO LOCK XXX from signature
	/* if not exist error */
	return ret;
}

int	validateSignature(XPGP *candidate, XPGP_SIGNATURE *signature, XPGP *validator)
{
	/* call from xPGP.h */
	EVP_PKEY *pubkey = NULL;
	int ret = 0;


	pubkey = XPGP_get_pubkey(validator);
	ret = XPGP_verify_it(candidate, signature, pubkey);

	EVP_PKEY_free(pubkey);
	//fprintf(stderr,"validateSignature() = %d\n", ret);
	return ret;

}

int	signCertificate(XPGP_KEYRING *kr, XPGP *subj, XPGP *issuer)
{
	fprintf(stderr,"signCertificate()\n");

	/* first check that we haven't already signed it */

	XPGP_SIGNATURE *sign;
	if (findSignature(subj, issuer, &sign))
	{
		fprintf(stderr,"signCertificate() Already signed by us!\n");
		return 1;
	}

	int ret = XPGP_sign_it(subj, issuer, kr->houseKeyPrivate, EVP_sha1());
	if (ret)
	{
		/* if we can remove it -> put in correct stack */
		if (removeCertificate(kr, subj))
		{
			addCertificateToStack(kr->trusted_ownsign, subj);
		}
		checkKeyRing(kr);
	}
	return ret;
}

// XXX must add keyid comparision.
//

int 	findSignature(XPGP *subj, XPGP *issuer, XPGP_SIGNATURE **sign)
{
        int i;
	XPGP_SIGNATURE *signature;

	fprintf(stderr,"findSignature()\n");

	for (i = 0; i < sk_XPGP_SIGNATURE_num(subj->signs); i++)
	{
                signature = sk_XPGP_SIGNATURE_value(subj->signs, i);
		if (0 == XPGP_SIGNER_cmp(issuer, signature))
		{
			*sign = signature;
			// SHOULD ADD CRYPTO LOCK XXX
			return 1;
		}
	}
	return 0;
}


 /**********************************************************************
  * Certificate Moving related Fns.
  **********************************************************************
  *
  * int 	existsCertificate(XPGP_KEYRING *kr, XPGP *key);
  * int 	addCertificateToStack(STACK_OF(XPGP) *sk, XPGP *unwanted);
  * int 	removeCertificateFromStack(STACK_OF(XPGP) *sk, XPGP *unwanted);
  * int 	existsCertificateInStack(STACK_OF(XPGP) *sk, XPGP *unwanted);
  * int 	moveStackToStack(STACK_OF(XPGP) *dest, STACK_OF(XPGP) *src);
  * 
  */


int 	existsCertificate(XPGP_KEYRING *kr, XPGP *key)
{
	/* check all stacks */
	int ret = 0;
	fprintf(stderr,"existsCertificate()\n");
	ret |= existsCertificateInStack(kr->unknown_certs, key);
	ret |= existsCertificateInStack(kr->untrusted_certs, key);
	ret |= existsCertificateInStack(kr->trusted_certs, key);
	ret |= existsCertificateInStack(kr->trusted_basic, key);
	ret |= existsCertificateInStack(kr->trusted_ownsign, key);
	ret |= existsCertificateInStack(kr->trusted_signers, key);
	return ret;
}

int 	addCertificateToStack(STACK_OF(XPGP) *sk, XPGP *newbie)
{
	fprintf(stderr,"addCertificateToStack()\n");

	// Believe this fn SHOULD AddE CRYPTO LOCK.
	fprintf(stderr,"addCertificateToStack() Pre-Push Ref: %d\n", newbie->references);
        sk_XPGP_push(sk, newbie);
	fprintf(stderr,"addCertificateToStack() Post-Push Ref: %d\n", newbie->references);

	return 1;
}


int 	removeCertificateFromStack(STACK_OF(XPGP) *sk, XPGP *unwanted)
{
	STACK_OF(XPGP) *topStack;
	int found = 0;
	XPGP *top = NULL;
        int i;

	// see if it exists first.
	if (!existsCertificateInStack(sk, unwanted))
	{
		fprintf(stderr,"removeCertificateFromStack() Fail - Not Exist\n");
		return 0;
	}

	topStack = sk_XPGP_new_null();
	for(i = 0; (0 < sk_XPGP_num(sk)) && (!found) ; i++)
	{
		top = sk_XPGP_pop(sk);
		// Bad comparision XXX
		if (top == unwanted)
		{
			found = 1;
		}
		else
		{
			// stick on tmp stack.
			sk_XPGP_push(topStack, top);
		}
	}

	// put the items back.
	while(0 < sk_XPGP_num(topStack))
	{
		top = sk_XPGP_pop(topStack);
		sk_XPGP_push(sk, top);
	}

	sk_XPGP_free(topStack);

	if (found)
	{
		fprintf(stderr,"removeCertificateFromStack() Removed\n");
	}
	else
	{
		fprintf(stderr,"removeCertificateFromStack() Fail - Not Exist2\n");
	}

	// done.
	return found;
}

int 	existsCertificateInStack(STACK_OF(XPGP) *sk, XPGP *toBorNot2B)
{
        int i;
	XPGP *xpgp;
	for (i = 0; i < sk_XPGP_num(sk); i++)
	{
                xpgp = sk_XPGP_value(sk, i);
		// Bad comparision XXX
		if (toBorNot2B == xpgp)
		{
			fprintf(stderr,"existsCertificateInStack() True\n");
			return 1;
		}
	}
	fprintf(stderr,"existsCertificateInStack() False\n");
	return 0;
}

int moveStackToStack(STACK_OF(XPGP) *dest, STACK_OF(XPGP) *src)
{
	int i;
	XPGP *xpgp;
	fprintf(stderr,"moveStackToStack()\n");
	for(i = 0; 0 < sk_XPGP_num(src); i++)
	{
		xpgp = sk_XPGP_pop(src);
		sk_XPGP_push(dest, xpgp);
	}
	return 1;
}

		

 /**********************************************************************
  * Validation Fns.
  **********************************************************************
  *
int validateOrClearCertifateStack(STACK_OF(XPGP) *);
int validateSelfSignedCertificate(XPGP *cert)
int checkOrClearStackSignedByKey(STACK_OF(XPGP) *sk, XPGP *pgp);
int validateCertificateIsSignedByKey(XPGP *subj, XPGP *issuer)

int checkStackTrustworthiness(XPGP_KEYRING *kr, STACK_OF(XPGP) *sk, int skLvl)
int moveToStack(SPGP_KEYRING *kr, XPGP *xpgp, int skLvl)
  *
  *
  */



int validateOrClearCertifateStack(STACK_OF(XPGP) *sk)
{
	STACK_OF(XPGP) *topStack = sk_XPGP_new_null();
	STACK_OF(XPGP) *rmStack = sk_XPGP_new_null();
	XPGP *top = NULL;
        int i;

	fprintf(stderr,"validateOrClearCertifateStack()\n");
	for(i = 0; (0 < sk_XPGP_num(sk)) ; i++)
	{
		top = sk_XPGP_pop(sk);
		if (validateSelfSignedCertificate(top))
		{
			sk_XPGP_push(topStack, top);
		}
		else
		{
			sk_XPGP_push(rmStack, top);
		}
	}

	// put the items back.
	while(0 < sk_XPGP_num(topStack))
	{
		top = sk_XPGP_pop(topStack);
		sk_XPGP_push(sk, top);
	}

	// clean up rmStack
	while(0 < sk_XPGP_num(topStack))
	{
		top = sk_XPGP_pop(topStack);
		//XPGP_free(top);
	}

	// clean up the rmStack.
	sk_XPGP_free(rmStack);
	sk_XPGP_free(topStack);

	return 1;
}

int validateSelfSignedCertificate(XPGP *cert)
{
	// get the first signature....
	// this, by convention should be the self-signature.
	fprintf(stderr,"validateSelfSignedCertificate()\n");
	if (sk_XPGP_SIGNATURE_num(cert->signs) < 1)
	{
		return 0;
	}
	XPGP_SIGNATURE *ss = sk_XPGP_SIGNATURE_value(cert->signs, 0);
	EVP_PKEY *key = XPGP_get_pubkey(cert);

	return XPGP_verify_it(cert, ss, key);
}


int checkOrClearStackSignedByKey(STACK_OF(XPGP) *sk, XPGP *pgp)
{
	STACK_OF(XPGP) *topStack = sk_XPGP_new_null();
	STACK_OF(XPGP) *rmStack = sk_XPGP_new_null();
	XPGP *top = NULL;
        int i;

	fprintf(stderr,"checkOrClearStackSignedByKey()\n");
	for(i = 0; (0 < sk_XPGP_num(sk)) ; i++)
	{
		top = sk_XPGP_pop(sk);
		if (validateCertificateIsSignedByKey(top, pgp))
		{
			sk_XPGP_push(topStack, top);
		}
		else
		{
			sk_XPGP_push(rmStack, top);
		}
	}

	// put the items back.
	while(0 < sk_XPGP_num(topStack))
	{
		top = sk_XPGP_pop(topStack);
		sk_XPGP_push(sk, top);
	}

	// put the items back.
	while(0 < sk_XPGP_num(rmStack))
	{
		top = sk_XPGP_pop(topStack);
		//XPGP_free(top);
	}

	// clean up the rmStack.
	sk_XPGP_free(rmStack);
	sk_XPGP_free(topStack);

	return 1;
}

int checkAndMoveSignedByOwnKey(XPGP_KEYRING *kr, STACK_OF(XPGP) *sk)
{
	STACK_OF(XPGP) *topStack = sk_XPGP_new_null();
	XPGP *top = NULL;
        int i;

	fprintf(stderr,"checkAndMoveSignedByOwnKey()\n");
	for(i = 0; (0 < sk_XPGP_num(sk)) ; i++)
	{
		top = sk_XPGP_pop(sk);
		if (validateCertificateIsSignedByKey(top, kr->houseKeyPublic))
		{
			sk_XPGP_push(kr->trusted_ownsign, top);
		}
		else
		{
			sk_XPGP_push(topStack, top);
		}
	}

	// put the items back.
	while(0 < sk_XPGP_num(topStack))
	{
		top = sk_XPGP_pop(topStack);
		sk_XPGP_push(sk, top);
	}

	// clean up the rmStack.
	sk_XPGP_free(topStack);

	return 1;
}

int validateCertificateIsSignedByKey(XPGP *subj, XPGP *issuer)
{
	/* loop through the signature */
	// get the first signature....
	// this, by convention should be the self-signature.
	XPGP_SIGNATURE *ss;
	EVP_PKEY *key = NULL;
	int found = 0;
	int i;

	fprintf(stderr,"validateCertificateIsSignedByKey()\n");
	if((!issuer) || (!subj))
	{
		return 0;
	}

	key = XPGP_get_pubkey(issuer);

	for(i = 0;( i < sk_XPGP_SIGNATURE_num(subj->signs)) && (!found); i++)
	{
		ss = sk_XPGP_SIGNATURE_value(subj->signs, i);
		if (0 == XPGP_SIGNER_cmp(issuer, ss))
		{

			if (XPGP_verify_it(subj, ss, key))
			{
				found = 1;
			}
		}
	}
	return found;
}

// Shouldn't be used on TRUSTED certs....
// as the trust level can conflict....
int checkStackTrustworthiness(XPGP_KEYRING *kr, STACK_OF(XPGP) *sk, int skLvl)
{
        int i;
	int moved = 0;
	int ret = 0;
	fprintf(stderr,"checkStackTrustworthiness()\n");
	for (i = 0; i < sk_XPGP_num(sk); i++)
	{
		XPGP *xpgp = sk_XPGP_value(sk, i);
		ret = getTrustLevel(kr, xpgp);
		if (ret != skLvl)
		{
			if (removeCertificateFromStack(sk, xpgp))
			{
				moveToStack(kr, xpgp, ret);
				moved++;
			}
			else
			{
				fprintf(stderr, "** ERROR cert: not in Stack:(%d->%d) : ",
					skLvl, ret);
				printX509NameString(stderr, XPGP_get_uid_name(xpgp));
				fprintf(stderr, "\n Removing Globally!\n");
				removeCertificate(kr, xpgp);
				moveToStack(kr, xpgp, ret);
			}
				
		}
	}
	return moved;
}


	
int moveToStack(XPGP_KEYRING *kr, XPGP *xpgp, int skLvl)
{
	STACK_OF(XPGP) *sk;

	fprintf(stderr,"moveToStack() %d: ", skLvl);
	printX509NameString(stderr, XPGP_get_uid_name(xpgp));
	fprintf(stderr,"\n");

	switch(skLvl)
	{
		case TRUST_SIGN_OWN:
			sk = kr->trusted_ownsign;
			addCertificateToStack(sk, xpgp);
			break;
		case TRUST_SIGN_TRSTED:
			sk = kr->trusted_signers;
			addCertificateToStack(sk, xpgp);
			break;
		case TRUST_SIGN_AUTHEN:
			sk = kr->trusted_certs;
			addCertificateToStack(sk, xpgp);
			break;
		case TRUST_SIGN_BASIC:
			sk = kr->trusted_basic;
			addCertificateToStack(sk, xpgp);
			break;
		case TRUST_SIGN_UNTRUSTED:
			sk = kr->untrusted_certs;
			addCertificateToStack(sk, xpgp);
			break;
		case TRUST_SIGN_UNKNOWN:
		case TRUST_SIGN_NONE:
		default:
			sk = kr->unknown_certs;
			addCertificateToStack(sk, xpgp);
			break;
	}
	return 1;
}

		
 /**********************************************************************
  * Trust Calculations +  Control Fns.
  **********************************************************************
  *
  * int     setSecurityLevel(XPGP_KEYRING *, int val)
  * int     getTrustLevel(XPGP_KEYRING *, XPGP *peer)
  * int	   getTrustSignatureLevel(XPGP_KEYRING *, XPGP *candidate, XPGP_SIGNATURE *sign)
  * XPGP   *findSigner(STACK_OF(XPGP) *sk, XPGP_SIGNATURE *sign)
  * int	    XPGP_SIGNER_cmp(XPGP *issuer, XPGP_SIGNATURE *sign)
  * int     validateCertificate(XPGP_KEYRING *ikr, XPGP *peer)
  *
  *
  */


int     setSecurityLevel(XPGP_KEYRING *kr, int val)
{
	fprintf(stderr,"setSecurityLevel()\n");
	kr->secLevel = val;
	if (val > 99)
	{
		val = 99;
	}
	else if (val < 1)
	{
		val = 1;
	}
	kr->state = XPGP_KEYRING_STATE_BAD;
	return 1;
}

int     getTrustLevel(XPGP_KEYRING *kr, XPGP *peer)
{

	/* iterate through the peers signatures 
	 * for each, check signature, and get score
	 */
	int trust_level_ownsign = 0;
	int trust_level_trusted = 0;
	int trust_level_authen  = 0;
	int trust_level_basic   = 0;

	int trust_count_ownsign = 0;
	int trust_count_trusted = 0;
	int trust_count_authen  = 0;
	int trust_count_basic   = 0;

	int trust_total = 0;
	int ret;

        int i;
	//fprintf(stderr,"getTrustLevel()\n");
	/*
	 * Skipping the first signature 
	 * (self) - which should already have been checked 
	 */

	for (i = 1; i < sk_XPGP_SIGNATURE_num(peer->signs); i++)
	{
		XPGP_SIGNATURE *sign = sk_XPGP_SIGNATURE_value(peer->signs, i);
		/* find certificate */
		ret = getTrustSignatureLevel(kr, peer, sign);

//		fprintf(stderr, "Trust Levels -------------------\n");
//		fprintf(stderr, "Subject: ");
//		printX509NameString(stderr, XPGP_get_uid_name(peer));
//		fprintf(stderr, "\n");
//		fprintf(stderr, "Issuer : ");
//		printX509NameString(stderr, 
//				XPGP_SIGNATURE_get_issuer_name(sign));
//		fprintf(stderr, "\n");
//		fprintf(stderr, "Trust Lvl: %d\n", ret);
//		fprintf(stderr, "Trust Levels -------------------\n");

		switch(ret)
		{
			case TRUST_SIGN_OWN:
				trust_count_ownsign++;
				break;
			case TRUST_SIGN_TRSTED:
				trust_count_trusted++;
				break;
			case TRUST_SIGN_AUTHEN:
				trust_count_authen++;
				break;
			case TRUST_SIGN_BASIC:
				trust_count_basic++;
				break;
			default:
				/* nothing! */
				break;
		}
	}

	/* Add up the scores */
	trust_level_ownsign = trust_count_ownsign * 
			(TRUST_SIGN_OWN_MIN + 
			((100.0 - kr->secLevel) / 100.0) * 
			(TRUST_SIGN_OWN_MAX - TRUST_SIGN_OWN_MIN));

	trust_level_trusted = trust_count_trusted *
			(TRUST_SIGN_TRSTED_MIN + 
			((100.0 - kr->secLevel) / 100.0) * 
			(TRUST_SIGN_TRSTED_MAX - TRUST_SIGN_TRSTED_MIN));

	trust_level_authen = trust_count_authen *
			(TRUST_SIGN_AUTHEN_MIN + 
			((100.0 - kr->secLevel) / 100.0) * 
			(TRUST_SIGN_AUTHEN_MAX - TRUST_SIGN_AUTHEN_MIN));

	trust_level_basic = trust_count_basic *
			(TRUST_SIGN_BASIC_MIN + (int)
			((100.0 - kr->secLevel) / 100.0) * 
			(TRUST_SIGN_BASIC_MAX - TRUST_SIGN_BASIC_MIN));

	/* total up there score */
	if (trust_level_basic + trust_level_authen > 40)
	{
		trust_total = 40;
	}
	else
	{
		trust_total = trust_level_basic + trust_level_authen;
	}

	trust_total += trust_level_trusted;
	trust_total += trust_level_ownsign;

	fprintf(stderr, "Trust Levels -------------------\n");

	printX509NameString(stderr, XPGP_get_uid_name(peer));
	fprintf(stderr, "\n");
	fprintf(stderr, "Trust Levels -------------------\n");
	fprintf(stderr, "                    Count Score\n");
	fprintf(stderr, "        Own Signing: %03d %04d\n", 
			trust_count_ownsign, trust_level_ownsign);
	fprintf(stderr, "     Trusted Signer: %03d %04d\n", 
			trust_count_trusted, trust_level_trusted);
	fprintf(stderr, " Authenticated Cert: %03d %04d\n", 
			trust_count_authen, trust_level_authen);
	fprintf(stderr, "     Marginal Trust: %03d %04d\n", 
			trust_count_basic, trust_level_basic);
	fprintf(stderr, "Trust Levels -------------------\n");

	// The two very trusted types.
	if (trust_count_ownsign > 0)
	{
		/* check if trusted signer! */
		if(existsCertificateInStack(kr->trusted_signers, peer))
		{
			return TRUST_SIGN_TRSTED;
		}
		return TRUST_SIGN_OWN;
	}

	if (trust_total <= 0)
	{
		/* should be in Unknown. */
		return TRUST_SIGN_UNKNOWN;
	}
	else if (trust_total < TRUST_BASIC)
	{
		return TRUST_SIGN_UNTRUSTED;
	}
	else if (trust_total < TRUST_AUTHEN)
	{
		return TRUST_SIGN_BASIC;
	}
	return TRUST_SIGN_AUTHEN;
}


int getTrustSignatureLevel(XPGP_KEYRING *kr, XPGP *candidate, XPGP_SIGNATURE *sign)
{

	/* so check if its in the trust_ownsign list */
	XPGP *signer = kr->houseKeyPublic;
	//fprintf(stderr,"getTrustSignatureLevel()\n");
	if (0 == XPGP_SIGNER_cmp(signer, sign))
	{
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_OWN;
		}
	}

	signer = findSigner(kr->trusted_signers, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in TS\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_TRSTED;
		}
		else
		{
		}
	}

	signer = findSigner(kr->trusted_ownsign, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in TO\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_AUTHEN;
		}
		else
		{
		}
	}

	signer = findSigner(kr->trusted_certs, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in TC\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_AUTHEN;
		}
		else
		{
		}
	}

	signer = findSigner(kr->trusted_basic, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in TB\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_BASIC;
		}
		else
		{
		}
	}

	signer = findSigner(kr->untrusted_certs, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in UT\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_UNTRUSTED;
		}
		else
		{
		}
	}

	signer = findSigner(kr->unknown_certs, sign);
	if (signer)
	{
		fprintf(stderr, "getTSL() found signer in UK\n");
		if (validateSignature(candidate, sign, signer))
		{
			return TRUST_SIGN_UNKNOWN;
		}
		else
		{
		}
	}
	//fprintf(stderr, "getTSL() signer unknown\n");
	return TRUST_SIGN_NONE;
}


XPGP   *findSigner(STACK_OF(XPGP) *sk, XPGP_SIGNATURE *sign)
{
        int i;
	//fprintf(stderr,"findSigner()\n");
	for (i = 0; i < sk_XPGP_num(sk); i++)
	{
		XPGP *xpgp = sk_XPGP_value(sk, i);
		if (0 == XPGP_SIGNER_cmp(xpgp, sign))
		{
			return xpgp;
		}
	}
	return NULL;
}

// XXX keyid should be used!!!.
// returns 0 == , -1 < . 1 >
int	XPGP_SIGNER_cmp(XPGP *issuer, XPGP_SIGNATURE *sign)
{
	int ret = X509_NAME_cmp(sign->issuer,issuer->subject->subject);
	//fprintf(stderr,"XPGP_SIGNER_cmp() = %d\n", ret);
	return ret;
}

int	XPGP_SIGNATURE_cmp(XPGP_SIGNATURE *s1, XPGP_SIGNATURE *s2)
{
	int ret = X509_NAME_cmp(s1->issuer,s2->issuer);
	if (ret != 0)
	{
		return ret;
	}

	/* check actual signature! */
	/* check len */

	/* check bytes */
	ASN1_BIT_STRING *sign1 = s1->signature;
	ASN1_BIT_STRING *sign2 = s2->signature;
	int signlen1 = ASN1_STRING_length(sign1);
	int signlen2 = ASN1_STRING_length(sign2);

	ret = signlen1 - signlen2;
	if (ret != 0)
	{
		return ret;
	}

	unsigned char *signdata1 = ASN1_STRING_data(sign1);
	unsigned char *signdata2 = ASN1_STRING_data(sign2);
	int j;
	for(j=0;j<signlen1;j++)
	{
		ret = (signdata1[j] - signdata2[j]);
		if (ret != 0)
		{
			return ret;
		}
	}
	return 0; // the same.
}


int     validateCertificate(XPGP_KEYRING *kr, XPGP *peer)
{
	//fprintf(stderr,"validateCertificate()\n");
	return getTrustLevel(kr, peer);
}
	

int printX509NameString(FILE *out, X509_NAME *name)
{
	int i;
	for(i = 0; i < X509_NAME_entry_count(name); i++)
	{
		X509_NAME_ENTRY *entry = X509_NAME_get_entry(name, i);
		ASN1_STRING *entry_data = X509_NAME_ENTRY_get_data(entry);
		//ASN1_OBJECT *entry_obj = X509_NAME_ENTRY_get_object(entry);
	
		if (entry_data -> data != NULL)
		{
			fprintf(out, ": %s ", (char *) entry_data -> data);
		}
	}
	return 1;
}


int 	DisplayStacks(XPGP_KEYRING *kr)
{
	fprintf(stderr, "DIsplayStacks()\n");
	fprintf(stderr, "Trusted Signers =====================\n");
	DisplayStack(kr->trusted_signers);
	fprintf(stderr, "Trusted OwnSign =====================\n");
	DisplayStack(kr->trusted_ownsign);
	fprintf(stderr, "Trusted Certs   =====================\n");
	DisplayStack(kr->trusted_certs);
	fprintf(stderr, "Trusted Basic   =====================\n");
	DisplayStack(kr->trusted_basic);
	fprintf(stderr, "UnTrusted Certs =====================\n");
	DisplayStack(kr->untrusted_certs);
	fprintf(stderr, "Unknown Certs   =====================\n");
	DisplayStack(kr->unknown_certs);

	return 1;
}

int 	DisplayStackTrust(XPGP_KEYRING *kr)
{
	checkStackTrustworthiness(kr, kr->unknown_certs, TRUST_SIGN_UNKNOWN);
	checkStackTrustworthiness(kr, kr->untrusted_certs, TRUST_SIGN_UNTRUSTED);
	checkStackTrustworthiness(kr, kr->trusted_basic, TRUST_SIGN_BASIC);
	checkStackTrustworthiness(kr, kr->trusted_certs, TRUST_SIGN_AUTHEN);
	checkStackTrustworthiness(kr, kr->trusted_signers, TRUST_SIGN_TRSTED);
	checkStackTrustworthiness(kr, kr->trusted_ownsign, TRUST_SIGN_OWN);
	return 1;
}

int DisplayStack(STACK_OF(XPGP) *sk)
{
	int i, j;
	for(i = 0; i < sk_XPGP_num(sk); i++)
	{
		XPGP *xpgp = sk_XPGP_value(sk,i);
		fprintf(stderr, "Cert : ");
		printX509NameString(stderr, XPGP_get_uid_name(xpgp));
		fprintf(stderr, "\n");

		for(j = 0; j < sk_XPGP_SIGNATURE_num(xpgp->signs); j++)
		{
			XPGP_SIGNATURE *sign = sk_XPGP_SIGNATURE_value(xpgp->signs,j);
			fprintf(stderr, "\tSigner : ");
			printX509NameString(stderr, 
				XPGP_SIGNATURE_get_issuer_name(sign));
			fprintf(stderr, "\n");
		}
	}
	return 1;
}


		

	
	
IMPLEMENT_STACK_OF(XPGP)
IMPLEMENT_ASN1_SET_OF(XPGP)

