/* ssl/ssl_rsa.c */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 * 
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 * 
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from 
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 * 
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */

#include <stdio.h>
#include "ssl_locl.h"
#include <openssl/bio.h>
#include <openssl/objects.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/pem.h>

#include "pgp1.h"

static int ssl_set_pgp_cert(CERT *c, XPGP *xpgp);
static int ssl_set_pgp_pkey(CERT *c, EVP_PKEY *pkey);

int SSL_use_pgp_certificate(SSL *ssl, XPGP *x)
	{

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_certificate()\n");
#endif /* XPGP_DEBUG */

	if (x == NULL)
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "SSL_use_pgp_certificate() ERR 1\n");
#endif /* XPGP_DEBUG */
		SSLerr(SSL_F_SSL_USE_CERTIFICATE,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ssl->cert))
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "SSL_use_pgp_certificate() ERR 2\n");
#endif /* XPGP_DEBUG */
		SSLerr(SSL_F_SSL_USE_CERTIFICATE,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	return(ssl_set_pgp_cert(ssl->cert,x));
	}

#ifndef OPENSSL_NO_STDIO
int SSL_use_pgp_certificate_file(SSL *ssl, const char *file, int type)
	{
	int j;
	BIO *in;
	int ret=0;
	XPGP *x=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_certificate_file()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_USE_CERTIFICATE_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_USE_CERTIFICATE_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if (type == SSL_FILETYPE_ASN1)
		{
		j=ERR_R_ASN1_LIB;
		x=d2i_XPGP_bio(in,NULL);
		}
	else if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		x=PEM_read_bio_XPGP(in,NULL,ssl->ctx->default_passwd_callback,ssl->ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_USE_CERTIFICATE_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}

	if (x == NULL)
		{
		SSLerr(SSL_F_SSL_USE_CERTIFICATE_FILE,j);
		goto end;
		}

	ret=SSL_use_pgp_certificate(ssl,x);
end:
	if (x != NULL) XPGP_free(x);
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif


#ifndef OPENSSL_NO_RSA
int SSL_use_pgp_RSAPrivateKey(SSL *ssl, RSA *rsa)
	{
	EVP_PKEY *pkey;
	int ret;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_RSAPrivateKey()\n");
#endif /* XPGP_DEBUG */

	if (rsa == NULL)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ssl->cert))
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	if ((pkey=EVP_PKEY_new()) == NULL)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY,ERR_R_EVP_LIB);
		return(0);
		}

	RSA_up_ref(rsa);
	EVP_PKEY_assign_RSA(pkey,rsa);

	ret=ssl_set_pgp_pkey(ssl->cert,pkey);
	EVP_PKEY_free(pkey);
	return(ret);
	}
#endif


static int ssl_set_pgp_pkey(CERT *c, EVP_PKEY *pkey)
	{
	int i,ok=0,bad=0;

#ifdef XPGP_DEBUG
	fprintf(stderr, "ssl_set_pgp_pkey()\n");
#endif /* XPGP_DEBUG */

	i=ssl_cert_type(NULL,pkey);
	if (i < 0)
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() ERROR bad type\n");
#endif /* XPGP_DEBUG */
		SSLerr(SSL_F_SSL_SET_PKEY,SSL_R_UNKNOWN_CERTIFICATE_TYPE);
		return(0);
		}

	// rmf24 extra check.... for correct type.
	if ((c->pkeys[i].gct.xpgp != NULL) && (c->pkeys[i].ctype != CERT_TYPE_XPGP))
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() ERROR not XPGP\n");
#endif /* XPGP_DEBUG */
		SSLerr(SSL_F_SSL_SET_PKEY,SSL_R_UNKNOWN_CERTIFICATE_TYPE);
		return(0);
		}

	if (c->pkeys[i].gct.xpgp != NULL)
		{
		EVP_PKEY *pktmp;
		pktmp =	XPGP_get_pubkey(c->pkeys[i].gct.xpgp);
		EVP_PKEY_copy_parameters(pktmp,pkey);
		EVP_PKEY_free(pktmp);
		ERR_clear_error();
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() XPGP != NULL\n");
#endif /* XPGP_DEBUG */

#ifndef OPENSSL_NO_RSA
		/* Don't check the public/private key, this is mostly
		 * for smart cards. */
		if ((pkey->type == EVP_PKEY_RSA) &&
			(RSA_flags(pkey->pkey.rsa) &
			 RSA_METHOD_FLAG_NO_CHECK))
		  {
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_pkey() FLAG_NON_CHK\n");
#endif /* XPGP_DEBUG */
			 ok=1;
		  }
		else
#endif
		     if (!XPGP_check_private_key(c->pkeys[i].gct.xpgp,pkey))
			{
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_pkey() chk privkey failed\n");
#endif /* XPGP_DEBUG */
			if ((i == SSL_PKEY_DH_RSA) || (i == SSL_PKEY_DH_DSA))
				{
#ifdef XPGP_DEBUG
				fprintf(stderr, "ssl_set_pgp_pkey() reset key type\n");
#endif /* XPGP_DEBUG */
				i=(i == SSL_PKEY_DH_RSA)?
					SSL_PKEY_DH_DSA:SSL_PKEY_DH_RSA;

				if (c->pkeys[i].gct.xpgp == NULL)
					{
#ifdef XPGP_DEBUG
				fprintf(stderr, "ssl_set_pgp_pkey() xpgp==NULL, ok\n");
#endif /* XPGP_DEBUG */
					ok=1;
					}
				else
					{
					if (!XPGP_check_private_key(
						c->pkeys[i].gct.xpgp,pkey))
						{
#ifdef XPGP_DEBUG
				fprintf(stderr, "ssl_set_pgp_pkey() chk privkey failed 2\n");
#endif /* XPGP_DEBUG */
						bad=1;
						}
					else
						{
#ifdef XPGP_DEBUG
				fprintf(stderr, "ssl_set_pgp_pkey() ok (strange)\n");
#endif /* XPGP_DEBUG */
						ok=1;
						}
					}
				}
			else
				{
#ifdef XPGP_DEBUG
				fprintf(stderr, "ssl_set_pgp_pkey() ->bad\n");
#endif /* XPGP_DEBUG */
				bad=1;
				}
			}
		else
		  {
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_pkey() ok chk passed\n");
#endif /* XPGP_DEBUG */
			ok=1;
		  }
		}
	else
	{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() XPGP == NULL, ok\n");
#endif /* XPGP_DEBUG */
		ok=1;
	}

	if (bad)
		{
		XPGP_free(c->pkeys[i].gct.xpgp);
		c->pkeys[i].gct.xpgp=NULL;
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() freed old (bad) cert\n");
		fprintf(stderr, "ssl_set_pgp_pkey() returning failure\n");
#endif /* XPGP_DEBUG */
		return(0);
		}

	ERR_clear_error(); /* make sure no error from X509_check_private_key()
	                    * is left if we have chosen to ignore it */
	if (c->pkeys[i].privatekey != NULL)
	{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_pkey() freeing old privkey\n");
#endif /* XPGP_DEBUG */
		EVP_PKEY_free(c->pkeys[i].privatekey);
	}

#ifdef XPGP_DEBUG
	fprintf(stderr, "ssl_set_pgp_pkey() setting new privkey\n");
#endif /* XPGP_DEBUG */
	CRYPTO_add(&pkey->references,1,CRYPTO_LOCK_EVP_PKEY);
	c->pkeys[i].privatekey=pkey;
	c->key= &(c->pkeys[i]);

	c->valid=0;
	return(1);
	}

#ifndef OPENSSL_NO_RSA
#ifndef OPENSSL_NO_STDIO
int SSL_use_pgp_RSAPrivateKey_file(SSL *ssl, const char *file, int type)
	{
	int j,ret=0;
	BIO *in;
	RSA *rsa=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_RSAPrivateKey()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if	(type == SSL_FILETYPE_ASN1)
		{
		j=ERR_R_ASN1_LIB;
		rsa=d2i_RSAPrivateKey_bio(in,NULL);
		}
	else if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		rsa=PEM_read_bio_RSAPrivateKey(in,NULL,
			ssl->ctx->default_passwd_callback,ssl->ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}
	if (rsa == NULL)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY_FILE,j);
		goto end;
		}
	ret=SSL_use_pgp_RSAPrivateKey(ssl,rsa);
	RSA_free(rsa);
end:
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif

int SSL_use_pgp_RSAPrivateKey_ASN1(SSL *ssl, unsigned char *d, long len)
	{
	int ret;
	const unsigned char *p;
	RSA *rsa;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_RSAPrivateKey_ASN1()\n");
#endif /* XPGP_DEBUG */

	p=d;
	if ((rsa=d2i_RSAPrivateKey(NULL,&p,(long)len)) == NULL)
		{
		SSLerr(SSL_F_SSL_USE_RSAPRIVATEKEY_ASN1,ERR_R_ASN1_LIB);
		return(0);
		}

	ret=SSL_use_pgp_RSAPrivateKey(ssl,rsa);
	RSA_free(rsa);
	return(ret);
	}
#endif /* !OPENSSL_NO_RSA */

int SSL_use_pgp_PrivateKey(SSL *ssl, EVP_PKEY *pkey)
	{
	int ret;

	if (pkey == NULL)
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ssl->cert))
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	ret=ssl_set_pgp_pkey(ssl->cert,pkey);
	return(ret);
	}

#ifndef OPENSSL_NO_STDIO
int SSL_use_pgp_PrivateKey_file(SSL *ssl, const char *file, int type)
	{
	int j,ret=0;
	BIO *in;
	EVP_PKEY *pkey=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_use_pgp_PrivateKey_file()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		pkey=PEM_read_bio_PrivateKey(in,NULL,
			ssl->ctx->default_passwd_callback,ssl->ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}
	if (pkey == NULL)
		{
		SSLerr(SSL_F_SSL_USE_PRIVATEKEY_FILE,j);
		goto end;
		}
	ret=SSL_use_pgp_PrivateKey(ssl,pkey);
	EVP_PKEY_free(pkey);
end:
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif

int SSL_CTX_use_pgp_certificate(SSL_CTX *ctx, XPGP *x)
	{

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_certificate()\n");
#endif /* XPGP_DEBUG */

	if (x == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ctx->cert))
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	return(ssl_set_pgp_cert(ctx->cert, x));
	}

static int ssl_set_pgp_cert(CERT *c, XPGP *x)
	{
	EVP_PKEY *pkey;
	int i,ok=0,bad=0;

#ifdef XPGP_DEBUG
	fprintf(stderr, "ssl_set_pgp_cert()\n");
#endif /* XPGP_DEBUG */

	pkey=XPGP_get_pubkey(x);
	if (pkey == NULL)
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_cert() ERR: Null key\n");
#endif /* XPGP_DEBUG */
		SSLerr(SSL_F_SSL_SET_CERT,SSL_R_X509_LIB);
		return(0);
		}

	i=ssl_pgp_cert_type(x,pkey);
#ifdef XPGP_DEBUG
	fprintf(stderr, "ssl_set_pgp_cert() key type: %d\n", i);
#endif /* XPGP_DEBUG */
	if (i < 0)
		{
		SSLerr(SSL_F_SSL_SET_CERT,SSL_R_UNKNOWN_CERTIFICATE_TYPE);
		EVP_PKEY_free(pkey);
		return(0);
		}

	if (c->pkeys[i].privatekey != NULL)
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_cert() c->pkeys[i].privatekey != NULL (copy)\n");
#endif /* XPGP_DEBUG */
		EVP_PKEY_copy_parameters(pkey,c->pkeys[i].privatekey);
		ERR_clear_error();

#ifndef OPENSSL_NO_RSA
		/* Don't check the public/private key, this is mostly
		 * for smart cards. */
		if ((c->pkeys[i].privatekey->type == EVP_PKEY_RSA) &&
			(RSA_flags(c->pkeys[i].privatekey->pkey.rsa) &
			 RSA_METHOD_FLAG_NO_CHECK))
		{
			 ok=1;
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_cert() RSA_NOCHK -> ok=1\n");
#endif /* XPGP_DEBUG */
		}
		else
#endif
		{
		if (!XPGP_check_private_key(x,c->pkeys[i].privatekey))
			{
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_cert() chk_pk_key failed!\n");
#endif /* XPGP_DEBUG */
			if ((i == SSL_PKEY_DH_RSA) || (i == SSL_PKEY_DH_DSA))
				{
				i=(i == SSL_PKEY_DH_RSA)?
					SSL_PKEY_DH_DSA:SSL_PKEY_DH_RSA;

				if (c->pkeys[i].privatekey == NULL)
					ok=1;
				else
					{
					if (!XPGP_check_private_key(x,
						c->pkeys[i].privatekey))
						bad=1;
					else
						ok=1;
					}
				}
			else
				bad=1;
			}
		else
			{
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_cert() chk_pk_key Good!\n");
#endif /* XPGP_DEBUG */
			ok=1;
			}
		} /* OPENSSL_NO_RSA */
		}
	else
		ok=1;

	EVP_PKEY_free(pkey);
	if (bad)
		{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_cert() deleting privkey\n");
#endif /* XPGP_DEBUG */
		EVP_PKEY_free(c->pkeys[i].privatekey);
		c->pkeys[i].privatekey=NULL;
		}

	if (c->pkeys[i].gct.xpgp != NULL)
	{
#ifdef XPGP_DEBUG
		fprintf(stderr, "ssl_set_pgp_cert() freeing old key type:%d\n", 
				c->pkeys[i].ctype);
#endif /* XPGP_DEBUG */

        	if (c->pkeys[i].ctype == CERT_TYPE_XPGP)
		{
			XPGP_free(c->pkeys[i].gct.xpgp);
		}
		else if (c->pkeys[i].ctype == CERT_TYPE_X509)
		{
			X509_free(c->pkeys[i].gct.x509);
		}
		else
		{
#ifdef XPGP_DEBUG
			fprintf(stderr, "ssl_set_pgp_cert() unknown old key type\n");
#endif /* XPGP_DEBUG */
		}
	}

	CRYPTO_add(&x->references,1,CRYPTO_LOCK_X509);

	c->pkeys[i].ctype=CERT_TYPE_XPGP;
	c->pkeys[i].gct.xpgp=x;

	c->key= &(c->pkeys[i]);

	c->valid=0;
	return(1);
	}

#ifndef OPENSSL_NO_STDIO
int SSL_CTX_use_pgp_certificate_file(SSL_CTX *ctx, const char *file, int type)
	{
	int j;
	BIO *in;
	int ret=0;
	XPGP *x=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_certificate_file()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if (type == SSL_FILETYPE_ASN1)
		{
		j=ERR_R_ASN1_LIB;
		x=d2i_XPGP_bio(in,NULL);
		}
	else if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		x=PEM_read_bio_XPGP(in,NULL,ctx->default_passwd_callback,ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}

	if (x == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE_FILE,j);
		goto end;
		}

	ret=SSL_CTX_use_pgp_certificate(ctx,x);
end:
	if (x != NULL) XPGP_free(x);
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif

int SSL_CTX_use_pgp_certificate_ASN1(SSL_CTX *ctx, int len, unsigned char *d)
	{
	XPGP *x;
	int ret;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_certificate_ASN1()\n");
#endif /* XPGP_DEBUG */

	x=d2i_XPGP(NULL,&d,(long)len);
	if (x == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_CERTIFICATE_ASN1,ERR_R_ASN1_LIB);
		return(0);
		}

	ret=SSL_CTX_use_pgp_certificate(ctx,x);
	XPGP_free(x);
	return(ret);
	}

#ifndef OPENSSL_NO_RSA
int SSL_CTX_use_pgp_RSAPrivateKey(SSL_CTX *ctx, RSA *rsa)
	{
	int ret;
	EVP_PKEY *pkey;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_RSAPrivateKey()\n");
#endif /* XPGP_DEBUG */

	if (rsa == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ctx->cert))
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	if ((pkey=EVP_PKEY_new()) == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY,ERR_R_EVP_LIB);
		return(0);
		}

	RSA_up_ref(rsa);
	EVP_PKEY_assign_RSA(pkey,rsa);

	ret=ssl_set_pgp_pkey(ctx->cert, pkey);
	EVP_PKEY_free(pkey);
	return(ret);
	}

#ifndef OPENSSL_NO_STDIO
int SSL_CTX_use_pgp_RSAPrivateKey_file(SSL_CTX *ctx, const char *file, int type)
	{
	int j,ret=0;
	BIO *in;
	RSA *rsa=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_RSAPrivateKey_file()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if	(type == SSL_FILETYPE_ASN1)
		{
		j=ERR_R_ASN1_LIB;
		rsa=d2i_RSAPrivateKey_bio(in,NULL);
		}
	else if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		rsa=PEM_read_bio_RSAPrivateKey(in,NULL,
			ctx->default_passwd_callback,ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}
	if (rsa == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY_FILE,j);
		goto end;
		}
	ret=SSL_CTX_use_pgp_RSAPrivateKey(ctx,rsa);
	RSA_free(rsa);
end:
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif

int SSL_CTX_use_pgp_RSAPrivateKey_ASN1(SSL_CTX *ctx, unsigned char *d, long len)
	{
	int ret;
	const unsigned char *p;
	RSA *rsa;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_RSAPrivateKey_ASN1()\n");
#endif /* XPGP_DEBUG */

	p=d;
	if ((rsa=d2i_RSAPrivateKey(NULL,&p,(long)len)) == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_RSAPRIVATEKEY_ASN1,ERR_R_ASN1_LIB);
		return(0);
		}

	ret=SSL_CTX_use_pgp_RSAPrivateKey(ctx,rsa);
	RSA_free(rsa);
	return(ret);
	}
#endif /* !OPENSSL_NO_RSA */

int SSL_CTX_use_pgp_PrivateKey(SSL_CTX *ctx, EVP_PKEY *pkey)
	{

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_PrivateKey()\n");
#endif /* XPGP_DEBUG */

	if (pkey == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY,ERR_R_PASSED_NULL_PARAMETER);
		return(0);
		}
	if (!ssl_cert_inst(&ctx->cert))
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY,ERR_R_MALLOC_FAILURE);
		return(0);
		}
	return(ssl_set_pgp_pkey(ctx->cert,pkey));
	}

#ifndef OPENSSL_NO_STDIO
int SSL_CTX_use_pgp_PrivateKey_file(SSL_CTX *ctx, const char *file, int type)
	{
	int j,ret=0;
	BIO *in;
	EVP_PKEY *pkey=NULL;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_PrivateKey_file()\n");
#endif /* XPGP_DEBUG */

	in=BIO_new(BIO_s_file_internal());
	if (in == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY_FILE,ERR_R_BUF_LIB);
		goto end;
		}

	if (BIO_read_filename(in,file) <= 0)
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY_FILE,ERR_R_SYS_LIB);
		goto end;
		}
	if (type == SSL_FILETYPE_PEM)
		{
		j=ERR_R_PEM_LIB;
		pkey=PEM_read_bio_PrivateKey(in,NULL,
			ctx->default_passwd_callback,ctx->default_passwd_callback_userdata);
		}
	else
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY_FILE,SSL_R_BAD_SSL_FILETYPE);
		goto end;
		}
	if (pkey == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY_FILE,j);
		goto end;
		}
	ret=SSL_CTX_use_pgp_PrivateKey(ctx,pkey);
	EVP_PKEY_free(pkey);
end:
	if (in != NULL) BIO_free(in);
	return(ret);
	}
#endif

int SSL_CTX_use_pgp_PrivateKey_ASN1(int type, SSL_CTX *ctx, unsigned char *d,
	     long len)
	{
	int ret;
	unsigned char *p;
	EVP_PKEY *pkey;

#ifdef XPGP_DEBUG
	fprintf(stderr, "SSL_CTX_use_pgp_PrivateKey_ASN1()\n");
#endif /* XPGP_DEBUG */

	p=d;
	if ((pkey=d2i_PrivateKey(type,NULL,&p,(long)len)) == NULL)
		{
		SSLerr(SSL_F_SSL_CTX_USE_PRIVATEKEY_ASN1,ERR_R_ASN1_LIB);
		return(0);
		}

	ret=SSL_CTX_use_pgp_PrivateKey(ctx,pkey);
	EVP_PKEY_free(pkey);
	return(ret);
	}



